"""
Manage Route53 records with Boto 3

.. versionadded:: 2017.7.0

Create and delete Route53 records. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses ``boto3``, which can be installed via package, or pip.

This module accepts explicit route53 credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    route53.keyid: GKTADJGHEIQSXMKKRBJ08H
    route53.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile, either
passed in as a dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
      keyid: GKTADJGHEIQSXMKKRBJ08H
      key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
      region: us-east-1

.. code-block:: yaml

    An exciting new AWS Route 53 Hosted Zone:
      boto_route53.hosted_zone_present:
        - Name: example.com.
        - PrivateZone: true
        - VPCs:
          - VPCName: MyLittleVPC
            VPCRegion: us-east-1
          - VPCId: vpc-12345678
        - region: us-east-1
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    mycnamerecord:
      boto_route53.rr_present:
        - Name: test.example.com.
        - ResourceRecords:
          - my-elb.us-east-1.elb.amazonaws.com.
        - DomainName: example.com.
        - TTL: 60
        - Type: CNAME
        - region: us-east-1
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

"""

# keep lint from choking
# pylint: disable=W0106


import logging
import uuid

import salt.utils.data
import salt.utils.dictupdate
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto3_route53.find_hosted_zone" in __salt__:
        return "boto3_route53"
    return (False, "boto3_route53 module could not be loaded")


def hosted_zone_present(
    name,
    Name=None,
    PrivateZone=False,
    CallerReference=None,
    Comment=None,
    VPCs=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure a hosted zone exists with the given attributes.

    name
        The name of the state definition.

    Name
        The name of the domain. This should be a fully-specified domain, and should terminate with a
        period. This is the name you have registered with your DNS registrar. It is also the name
        you will delegate from your registrar to the Amazon Route 53 delegation servers returned in
        response to this request.  If not provided, the value of name will be used.

    PrivateZone
        Set True if creating a private hosted zone.  If true, then 'VPCs' is also required.

    Comment
        Any comments you want to include about the hosted zone.

    CallerReference
        A unique string that identifies the request and that allows create_hosted_zone() calls to be
        retried without the risk of executing the operation twice.  This helps ensure idempotency
        across state calls, but can cause issues if a zone is deleted and then an attempt is made
        to recreate it with the same CallerReference.  If not provided, a unique UUID will be
        generated at each state run, which can potentially lead to duplicate zones being created if
        the state is run again while the previous zone creation is still in PENDING status (which
        can occasionally take several minutes to clear).  Maximum length of 128.

    VPCs
        A list of dicts, each dict composed of a VPCRegion, and either a VPCId or a VPCName.
        Note that this param is ONLY used if PrivateZone == True

        VPCId
            When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
            required.  Exclusive with VPCName.

        VPCName
            When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
            required.  Exclusive with VPCId.

        VPCRegion
            When creating a private hosted zone, the region of the associated VPC is required.  If
            not provided, an effort will be made to determine it from VPCId or VPCName, if
            possible.  This will fail if a given VPCName exists in multiple regions visible to the
            bound account, in which case you'll need to provide an explicit value for VPCRegion.
    """
    Name = Name if Name else name

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if not PrivateZone and VPCs:
        raise SaltInvocationError(
            "Parameter 'VPCs' is invalid when creating a public zone."
        )
    if PrivateZone and not VPCs:
        raise SaltInvocationError(
            "Parameter 'VPCs' is required when creating a private zone."
        )
    if VPCs:
        if not isinstance(VPCs, list):
            raise SaltInvocationError("Parameter 'VPCs' must be a list of dicts.")
        for v in VPCs:
            if not isinstance(v, dict) or not salt.utils.data.exactly_one(
                (v.get("VPCId"), v.get("VPCName"))
            ):
                raise SaltInvocationError(
                    "Parameter 'VPCs' must be a list of dicts, each composed "
                    "of either a 'VPCId' or a 'VPCName', and optionally a "
                    "'VPCRegion', to help distinguish between multitple matches."
                )
    # Massage VPCs into something AWS will accept...
    fixed_vpcs = []
    if PrivateZone:
        for v in VPCs:
            VPCId = v.get("VPCId")
            VPCName = v.get("VPCName")
            VPCRegion = v.get("VPCRegion")
            VPCs = __salt__["boto_vpc.describe_vpcs"](
                vpc_id=VPCId,
                name=VPCName,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            ).get("vpcs", [])
            if VPCRegion and VPCs:
                VPCs = [v for v in VPCs if v["region"] == VPCRegion]
            if not VPCs:
                ret["comment"] = (
                    "A VPC matching given criteria (vpc: {} / vpc_region: {}) not "
                    "found.".format(VPCName or VPCId, VPCRegion)
                )
                log.error(ret["comment"])
                ret["result"] = False
                return ret
            if len(VPCs) > 1:
                ret["comment"] = (
                    "Multiple VPCs matching given criteria (vpc: {} / vpc_region: "
                    "{}) found: {}.".format(
                        VPCName or VPCId, VPCRegion, ", ".join([v["id"] for v in VPCs])
                    )
                )
                log.error(ret["comment"])
                ret["result"] = False
                return ret
            vpc = VPCs[0]
            if VPCName:
                VPCId = vpc["id"]
            if not VPCRegion:
                VPCRegion = vpc["region"]
            fixed_vpcs += [{"VPCId": VPCId, "VPCRegion": VPCRegion}]

    create = False
    update_comment = False
    add_vpcs = []
    del_vpcs = []
    args = {
        "Name": Name,
        "PrivateZone": PrivateZone,
        "region": region,
        "key": key,
        "keyid": keyid,
        "profile": profile,
    }
    zone = __salt__["boto3_route53.find_hosted_zone"](**args)
    if not zone:
        create = True
        # Grrrr - can only pass one VPC when initially creating a private zone...
        # The rest have to be added (one-by-one) later in a separate step.
        if len(fixed_vpcs) > 1:
            add_vpcs = fixed_vpcs[1:]
            fixed_vpcs = fixed_vpcs[:1]
        CallerReference = CallerReference if CallerReference else str(uuid.uuid4())
    else:
        # Currently the only modifiable traits about a zone are associated VPCs and the comment.
        zone = zone[0]
        if PrivateZone:
            for z in zone.get("VPCs"):
                if z not in fixed_vpcs:
                    del_vpcs += [z]
            for z in fixed_vpcs:
                if z not in zone.get("VPCs"):
                    add_vpcs += [z]
        if zone["HostedZone"]["Config"].get("Comment") != Comment:
            update_comment = True

    if not (create or add_vpcs or del_vpcs or update_comment):
        ret["comment"] = f"Hostd Zone {Name} already in desired state"
        return ret

    if create:
        if __opts__["test"]:
            ret["comment"] = "Route 53 {} hosted zone {} would be created.".format(
                "private" if PrivateZone else "public", Name
            )
            ret["result"] = None
            return ret
        vpc_id = fixed_vpcs[0].get("VPCId") if fixed_vpcs else None
        vpc_region = fixed_vpcs[0].get("VPCRegion") if fixed_vpcs else None
        newzone = __salt__["boto3_route53.create_hosted_zone"](
            Name=Name,
            CallerReference=CallerReference,
            Comment=Comment,
            PrivateZone=PrivateZone,
            VPCId=vpc_id,
            VPCRegion=vpc_region,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        if newzone:
            newzone = newzone[0]
            ret["comment"] = "Route 53 {} hosted zone {} successfully created".format(
                "private" if PrivateZone else "public", Name
            )
            log.info(ret["comment"])
            ret["changes"]["new"] = newzone
        else:
            ret["comment"] = "Creation of Route 53 {} hosted zone {} failed".format(
                "private" if PrivateZone else "public", Name
            )
            log.error(ret["comment"])
            ret["result"] = False
            return ret

    if update_comment:
        if __opts__["test"]:
            ret["comment"] = (
                "Route 53 {} hosted zone {} comment would be updated.".format(
                    "private" if PrivateZone else "public", Name
                )
            )
            ret["result"] = None
            return ret
        r = __salt__["boto3_route53.update_hosted_zone_comment"](
            Name=Name,
            Comment=Comment,
            PrivateZone=PrivateZone,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        if r:
            r = r[0]
            msg = "Route 53 {} hosted zone {} comment successfully updated".format(
                "private" if PrivateZone else "public", Name
            )
            log.info(msg)
            ret["comment"] = "  ".join([ret["comment"], msg])
            ret["changes"]["old"] = zone
            ret["changes"]["new"] = salt.utils.dictupdate.update(
                ret["changes"].get("new", {}), r
            )
        else:
            ret["comment"] = (
                "Update of Route 53 {} hosted zone {} comment failed".format(
                    "private" if PrivateZone else "public", Name
                )
            )
            log.error(ret["comment"])
            ret["result"] = False
            return ret

    if add_vpcs or del_vpcs:
        if __opts__["test"]:
            ret["comment"] = (
                "Route 53 {} hosted zone {} associated VPCs would be updated.".format(
                    "private" if PrivateZone else "public", Name
                )
            )
            ret["result"] = None
            return ret
        all_added = True
        all_deled = True
        for (
            vpc
        ) in add_vpcs:  # Add any new first to avoid the "can't delete last VPC" errors.
            r = __salt__["boto3_route53.associate_vpc_with_hosted_zone"](
                Name=Name,
                VPCId=vpc["VPCId"],
                VPCRegion=vpc["VPCRegion"],
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not r:
                all_added = False
        for vpc in del_vpcs:
            r = __salt__["boto3_route53.disassociate_vpc_from_hosted_zone"](
                Name=Name,
                VPCId=vpc["VPCId"],
                VPCRegion=vpc["VPCRegion"],
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not r:
                all_deled = False

        ret["changes"]["old"] = zone
        ret["changes"]["new"] = __salt__["boto3_route53.find_hosted_zone"](**args)
        if all_added and all_deled:
            msg = "Route 53 {} hosted zone {} associated VPCs successfully updated".format(
                "private" if PrivateZone else "public", Name
            )
            log.info(msg)
            ret["comment"] = "  ".join([ret["comment"], msg])
        else:
            ret["comment"] = (
                "Update of Route 53 {} hosted zone {} associated VPCs failed".format(
                    "private" if PrivateZone else "public", Name
                )
            )
            log.error(ret["comment"])
            ret["result"] = False
            return ret

    return ret


def hosted_zone_absent(
    name, Name=None, PrivateZone=False, region=None, key=None, keyid=None, profile=None
):
    """
    Ensure the Route53 Hostes Zone described is absent

    name
        The name of the state definition.

    Name
        The name of the domain. This should be a fully-specified domain, and should terminate with a
        period.  If not provided, the value of name will be used.

    PrivateZone
        Set True if deleting a private hosted zone.

    """
    Name = Name if Name else name

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    args = {
        "Name": Name,
        "PrivateZone": PrivateZone,
        "region": region,
        "key": key,
        "keyid": keyid,
        "profile": profile,
    }
    zone = __salt__["boto3_route53.find_hosted_zone"](**args)
    if not zone:
        ret["comment"] = "Route 53 {} hosted zone {} already absent".format(
            "private" if PrivateZone else "public", Name
        )
        log.info(ret["comment"])
        return ret
    if __opts__["test"]:
        ret["comment"] = "Route 53 {} hosted zone {} would be deleted".format(
            "private" if PrivateZone else "public", Name
        )
        ret["result"] = None
        return ret
    zone = zone[0]
    Id = zone["HostedZone"]["Id"]
    if __salt__["boto3_route53.delete_hosted_zone"](
        Id=Id, region=region, key=key, keyid=keyid, profile=profile
    ):
        ret["comment"] = "Route 53 {} hosted zone {} deleted".format(
            "private" if PrivateZone else "public", Name
        )
        log.info(ret["comment"])
        ret["changes"]["old"] = zone
        ret["changes"]["new"] = None
    else:
        ret["comment"] = "Failed to delete Route 53 {} hosted zone {}".format(
            "private" if PrivateZone else "public", Name
        )
        log.info(ret["comment"])
        ret["result"] = False
        return ret

    return ret


def rr_present(
    name,
    HostedZoneId=None,
    DomainName=None,
    PrivateZone=False,
    Name=None,
    Type=None,
    SetIdentifier=None,
    Weight=None,
    Region=None,
    GeoLocation=None,
    Failover=None,
    TTL=None,
    ResourceRecords=None,
    AliasTarget=None,
    HealthCheckId=None,
    TrafficPolicyInstanceId=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure the Route53 record is present.

    name
        The name of the state definition.  This will be used for Name if the latter is
        not provided.

    HostedZoneId
        The ID of a zone to create the record in.  Exclusive with DomainName.

    DomainName
        The domain name of a zone to create the record in.  Exclusive with HostedZoneId.

    PrivateZone
        Set to True if the resource record should be in a private zone, False if public.

    Name
        Name of the Route 53 resource record being managed.

    Type
        The record type (A, NS, MX, TXT, etc.)

    SetIdentifier
        Valid for Weighted, Latency, Geolocation, and Failover resource record sets only.
        An identifier that differentiates among multiple resource record sets that have the same
        combination of DNS name and type.  The value of SetIdentifier must be unique for each
        resource record set that has the same combination of DNS name and type. Omit SetIdentifier
        for any other types of record sets.

    Weight
        Valid for Weighted resource record sets only.  Among resource record sets that have the
        same combination of DNS name and type, a value that determines the proportion of DNS
        queries that Amazon Route 53 responds to using the current resource record set. Amazon Route
        53 calculates the sum of the weights for the resource record sets that have the same
        combination of DNS name and type. Amazon Route 53 then responds to queries based on the
        ratio of a resource's weight to the total.

        Note the following:

        - You must specify a value for the Weight element for every weighted resource record set.
        - You can only specify one ResourceRecord per weighted resource record set.
        - You can't create latency, failover, or geolocation resource record sets that have the
          same values for the Name and Type elements as weighted resource record sets.
        - You can create a maximum of 100 weighted resource record sets that have the same values
          for the Name and Type elements.
        - For weighted (but not weighted alias) resource record sets, if you set Weight to 0 for a
          resource record set, Amazon Route 53 never responds to queries with the applicable value
          for that resource record set.  However, if you set Weight to 0 for all resource record
          sets that have the same combination of DNS name and type, traffic is routed to all
          resources with equal probability.  The effect of setting Weight to 0 is different when
          you associate health checks with weighted resource record sets. For more information,
          see `Options for Configuring Amazon Route 53 Active-Active and Active-Passive Failover`__
          in the Amazon Route 53 Developer Guide.

          .. __: http://docs.aws.amazon.com/Route53/latest/DeveloperGuide/dns-failover-configuring-options.html

    Region
        Valid for Latency-based resource record sets only.  The Amazon EC2 Region where the resource
        that is specified in this resource record set resides. The resource typically is an AWS
        resource, such as an EC2 instance or an ELB load balancer, and is referred to by an IP
        address or a DNS domain name, depending on the record type.

    GeoLocation
        Geo location resource record sets only.  A dict that lets you control how Route 53 responds
        to DNS queries based on the geographic origin of the query.  For example, if you want all
        queries from Africa to be routed to a web server with an IP address of 192.0.2.111, create a
        resource record set with a Type of A and a ContinentCode of AF.

        .. code-block:: text

            ContinentCode
                The two-letter code for the continent.
                Valid values: AF | AN | AS | EU | OC | NA | SA
                Constraint: Specifying ContinentCode with either CountryCode or SubdivisionCode
                            returns an InvalidInput error.
            CountryCode
                The two-letter code for the country.
            SubdivisionCode
                The code for the subdivision, for example, a state in the United States or a
                province in Canada.

        Notes

        - Creating geolocation and geolocation alias resource record sets in private hosted zones
          is not supported.
        - If you create separate resource record sets for overlapping geographic regions (for
          example, one resource record set for a continent and one for a country on the same
          continent), priority goes to the smallest geographic region. This allows you to route
          most queries for a continent to one resource and to route queries for a country on that
          continent to a different resource.
        - You can't create two geolocation resource record sets that specify the same geographic
          location.
        - The value ``*`` in the CountryCode element matches all geographic locations that aren't
          specified in other geolocation resource record sets that have the same values for the
          Name and Type elements.
        - Geolocation works by mapping IP addresses to locations.  However, some IP addresses
          aren't mapped to geographic locations, so even if you create geolocation resource
          record sets that cover all seven continents, Amazon Route 53 will receive some DNS
          queries from locations that it can't identify.  We recommend that you
          create a resource record set for which the value of CountryCode is
          ``*``, which handles both queries that come from locations for which you
          haven't created geolocation resource record sets and queries from IP
          addresses that aren't mapped to a location.  If you don't create a ``*``
          resource record set, Amazon Route 53 returns a "no answer" response
          for queries from those locations.
        - You can't create non-geolocation resource record sets that have the same values for the
          Name and Type elements as geolocation resource record sets.

    TTL
        The resource record cache time to live (TTL), in seconds.
        Note the following:

        - If you're creating an alias resource record set, omit TTL. Amazon Route 53 uses the
          value of TTL for the alias target.
        - If you're associating this resource record set with a health check (if you're adding
          a HealthCheckId element), we recommend that you specify a TTL of 60 seconds or less so
          clients respond quickly to changes in health status.
        - All of the resource record sets in a group of weighted, latency, geolocation, or
          failover resource record sets must have the same value for TTL.
        - If a group of weighted resource record sets includes one or more weighted alias
          resource record sets for which the alias target is an ELB load balancer, we recommend
          that you specify a TTL of 60 seconds for all of the non-alias weighted resource record
          sets that have the same name and type. Values other than 60 seconds (the TTL for load
          balancers) will change the effect of the values that you specify for Weight.

    ResourceRecords
        A list, containing one or more values for the resource record.  No single value can exceed
        4,000 characters.  For details on how to format values for different record types, see
        `Supported DNS Resource Record Types`__ in the Amazon Route 53 Developer Guide.

        .. __: http://docs.aws.amazon.com/Route53/latest/DeveloperGuide/ResourceRecordTypes.html

        Note:  You can specify more than one value for all record types except CNAME and SOA.

        It is also possible to pass "magic" strings as resource record values.  This functionality
        can easily be extended, but for the moment supports the following:

            'magic:ec2_instance_tag:some_tag_name:some_string:some_instance_attr'

        This tells salt to lookup an EC2 instance with a tag 'some_tag_name' which has the value
        'some_string' and substitute the 'some_instance_attr' attribute of that instance as the
        resource record value being evaluated.

        This should work generally for any EC2 instance tags, as long as the instance attribute
        being fetched is available to getattr(instance, 'attribute') as seen in the code below.
        Anything else will most likely require this function to be extended to handle it.

        The canonical use-case for this (at least at our site) is to query the Name tag (which
        we always populate with the host's FQDN) to lookup the public or private IPs bound to the
        instance, so we can then automgically create Route 53 records for them.

    AliasTarget
        The rules governing how to define an AliasTarget for the various supported use-cases are
        obtuse beyond reason and attempting to paraphrase them (or even worse, cut-and-paste them
        in their entirety) would be silly and counterproductive.  If you need this feature, then
        Read The Fine Materials at the `Boto 3 Route 53 page`__ and/or the `AWS Route 53 docs`__
        and suss them for yourself - I sure won't claim to understand them partcularly well.

        .. __: http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets
        .. __: http://docs.aws.amazon.com/Route53/latest/APIReference/API_AliasTarget.html

    region
        The region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.
    """
    Name = Name if Name else name

    if Type is None:
        raise SaltInvocationError(
            "'Type' is a required parameter when adding or updating resource records."
        )
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    args = {
        "Id": HostedZoneId,
        "Name": DomainName,
        "PrivateZone": PrivateZone,
        "region": region,
        "key": key,
        "keyid": keyid,
        "profile": profile,
    }
    zone = __salt__["boto3_route53.find_hosted_zone"](**args)
    if not zone:
        ret["comment"] = "Route 53 {} hosted zone {} not found".format(
            "private" if PrivateZone else "public", DomainName
        )
        log.info(ret["comment"])
        return ret
    zone = zone[0]
    HostedZoneId = zone["HostedZone"]["Id"]

    # Convert any magic RR values to something AWS will understand, and otherwise clean them up.
    fixed_rrs = []
    if ResourceRecords:
        for rr in ResourceRecords:
            if rr.startswith("magic:"):
                fields = rr.split(":")
                if fields[1] == "ec2_instance_tag":
                    if len(fields) != 5:
                        log.warning(
                            "Invalid magic RR value seen: '%s'.  Passing as-is.", rr
                        )
                        fixed_rrs += [rr]
                        continue
                    tag_name = fields[2]
                    tag_value = fields[3]
                    instance_attr = fields[4]
                    good_states = (
                        "pending",
                        "rebooting",
                        "running",
                        "stopping",
                        "stopped",
                    )
                    r = __salt__["boto_ec2.find_instances"](
                        tags={tag_name: tag_value},
                        return_objs=True,
                        in_states=good_states,
                        region=region,
                        key=key,
                        keyid=keyid,
                        profile=profile,
                    )
                    if len(r) < 1:
                        ret["comment"] = (
                            "No EC2 instance with tag {} == {} found".format(
                                tag_name, tag_value
                            )
                        )
                        log.error(ret["comment"])
                        ret["result"] = False
                        return ret
                    if len(r) > 1:
                        ret["comment"] = (
                            "Multiple EC2 instances with tag {} == {} found".format(
                                tag_name, tag_value
                            )
                        )
                        log.error(ret["comment"])
                        ret["result"] = False
                        return ret
                    instance = r[0]
                    res = getattr(instance, instance_attr, None)
                    if res:
                        log.debug(
                            "Found %s %s for instance %s",
                            instance_attr,
                            res,
                            instance.id,
                        )
                        fixed_rrs += [__salt__["boto3_route53.aws_encode"](res)]
                    else:
                        ret["comment"] = "Attribute {} not found on instance {}".format(
                            instance_attr, instance.id
                        )
                        log.error(ret["comment"])
                        ret["result"] = False
                        return ret
                else:
                    ret["comment"] = (
                        "Unknown RR magic value seen: {}.  Please extend the "
                        "boto3_route53 state module to add support for your preferred "
                        "incantation.".format(fields[1])
                    )
                    log.error(ret["comment"])
                    ret["result"] = False
                    return ret
            else:
                # for TXT records the entry must be encapsulated in quotes as required by the API
                # this appears to be incredibly difficult with the jinja templating engine
                # so inject the quotations here to make a viable ChangeBatch
                if Type == "TXT":
                    rr = f'"{rr}"'
                fixed_rrs += [rr]
        ResourceRecords = [{"Value": rr} for rr in sorted(fixed_rrs)]

    recordsets = __salt__["boto3_route53.get_resource_records"](
        HostedZoneId=HostedZoneId,
        StartRecordName=Name,
        StartRecordType=Type,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )

    if SetIdentifier and recordsets:
        log.debug(
            "Filter recordsets %s by SetIdentifier %s.", recordsets, SetIdentifier
        )
        recordsets = [r for r in recordsets if r.get("SetIdentifier") == SetIdentifier]
        log.debug("Resulted in recordsets %s.", recordsets)

    create = False
    update = False
    updatable = [
        "SetIdentifier",
        "Weight",
        "Region",
        "GeoLocation",
        "Failover",
        "TTL",
        "AliasTarget",
        "HealthCheckId",
        "TrafficPolicyInstanceId",
    ]
    if not recordsets:
        create = True
        if __opts__["test"]:
            ret["comment"] = (
                "Route 53 resource record {} with type {} would be added.".format(
                    Name, Type
                )
            )
            ret["result"] = None
            return ret
    elif len(recordsets) > 1:
        ret["comment"] = "Given criteria matched more than one ResourceRecordSet."
        log.error(ret["comment"])
        ret["result"] = False
        return ret
    else:
        rrset = recordsets[0]
        for u in updatable:
            if locals().get(u) != rrset.get(u):
                update = True
                break
        if rrset.get("ResourceRecords") is not None:
            if ResourceRecords != sorted(
                rrset.get("ResourceRecords"), key=lambda x: x["Value"]
            ):
                update = True
        elif (AliasTarget is not None) and (rrset.get("AliasTarget") is not None):
            if sorted(AliasTarget) != sorted(rrset.get("AliasTarget")):
                update = True

    if not create and not update:
        ret["comment"] = (
            "Route 53 resource record {} with type {} is already in the desired state."
            "".format(Name, Type)
        )
        log.info(ret["comment"])
        return ret
    else:
        if __opts__["test"]:
            ret["comment"] = (
                "Route 53 resource record {} with type {} would be updated.".format(
                    Name, Type
                )
            )
            ret["result"] = None
            return ret
        ResourceRecordSet = {"Name": Name, "Type": Type}
        if ResourceRecords:
            ResourceRecordSet["ResourceRecords"] = ResourceRecords
        for u in updatable:
            if locals().get(u) or (locals().get(u) == 0):
                ResourceRecordSet.update({u: locals().get(u)})
            else:
                log.debug(
                    "Not updating ResourceRecordSet with local value: %s",
                    locals().get(u),
                )

        ChangeBatch = {
            "Changes": [{"Action": "UPSERT", "ResourceRecordSet": ResourceRecordSet}]
        }

        if __salt__["boto3_route53.change_resource_record_sets"](
            HostedZoneId=HostedZoneId,
            ChangeBatch=ChangeBatch,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ):
            ret["comment"] = "Route 53 resource record {} with type {} {}.".format(
                Name, Type, "created" if create else "updated"
            )
            log.info(ret["comment"])
            if create:
                ret["changes"]["old"] = None
            else:
                ret["changes"]["old"] = rrset
            ret["changes"]["new"] = ResourceRecordSet
        else:
            ret["comment"] = (
                "Failed to {} Route 53 resource record {} with type {}.".format(
                    "create" if create else "update", Name, Type
                )
            )
            log.error(ret["comment"])
            ret["result"] = False

    return ret


def rr_absent(
    name,
    HostedZoneId=None,
    DomainName=None,
    PrivateZone=False,
    Name=None,
    Type=None,
    SetIdentifier=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure the Route53 record is deleted.

    name
        The name of the state definition.  This will be used for Name if the latter is
        not provided.

    HostedZoneId
        The ID of the zone to delete the record from.  Exclusive with DomainName.

    DomainName
        The domain name of the zone to delete the record from.  Exclusive with HostedZoneId.

    PrivateZone
        Set to True if the RR to be removed is in a private zone, False if public.

    Name
        Name of the resource record.

    Type
        The record type (A, NS, MX, TXT, etc.)

    SetIdentifier
        Valid for Weighted, Latency, Geolocation, and Failover resource record sets only.
        An identifier that differentiates among multiple resource record sets that have the same
        combination of DNS name and type.  The value of SetIdentifier must be unique for each
        resource record set that has the same combination of DNS name and type. Omit SetIdentifier
        for any other types of record sets.

    region
        The region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.
    """
    Name = Name if Name else name

    if Type is None:
        raise SaltInvocationError(
            "'Type' is a required parameter when deleting resource records."
        )
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    args = {
        "Id": HostedZoneId,
        "Name": DomainName,
        "PrivateZone": PrivateZone,
        "region": region,
        "key": key,
        "keyid": keyid,
        "profile": profile,
    }
    zone = __salt__["boto3_route53.find_hosted_zone"](**args)
    if not zone:
        ret["comment"] = "Route 53 {} hosted zone {} not found".format(
            "private" if PrivateZone else "public", DomainName
        )
        log.info(ret["comment"])
        return ret
    zone = zone[0]
    HostedZoneId = zone["HostedZone"]["Id"]

    recordsets = __salt__["boto3_route53.get_resource_records"](
        HostedZoneId=HostedZoneId,
        StartRecordName=Name,
        StartRecordType=Type,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if SetIdentifier and recordsets:
        log.debug(
            "Filter recordsets %s by SetIdentifier %s.", recordsets, SetIdentifier
        )
        recordsets = [r for r in recordsets if r.get("SetIdentifier") == SetIdentifier]
        log.debug("Resulted in recordsets %s.", recordsets)
    if not recordsets:
        ret["comment"] = (
            "Route 53 resource record {} with type {} already absent.".format(
                Name, Type
            )
        )
        return ret
    elif len(recordsets) > 1:
        ret["comment"] = "Given criteria matched more than one ResourceRecordSet."
        log.error(ret["comment"])
        ret["result"] = False
        return ret
    ResourceRecordSet = recordsets[0]
    if __opts__["test"]:
        ret["comment"] = (
            "Route 53 resource record {} with type {} would be deleted.".format(
                Name, Type
            )
        )
        ret["result"] = None
        return ret

    ChangeBatch = {
        "Changes": [{"Action": "DELETE", "ResourceRecordSet": ResourceRecordSet}]
    }

    if __salt__["boto3_route53.change_resource_record_sets"](
        HostedZoneId=HostedZoneId,
        ChangeBatch=ChangeBatch,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    ):
        ret["comment"] = "Route 53 resource record {} with type {} deleted.".format(
            Name, Type
        )
        log.info(ret["comment"])
        ret["changes"]["old"] = ResourceRecordSet
        ret["changes"]["new"] = None
    else:
        ret["comment"] = (
            "Failed to delete Route 53 resource record {} with type {}.".format(
                Name, Type
            )
        )
        log.error(ret["comment"])
        ret["result"] = False

    return ret



"""
This runner is used only for test purposes and serves no production purpose
"""

import time


def arg(*args, **kwargs):
    """
    Output the given args and kwargs

    Kwargs will be filtered for 'private' keynames.

    CLI Example:

    .. code-block:: bash

        salt-run test.arg foo bar=baz
    """
    kwargs = {k: v for k, v in kwargs.items() if not k.startswith("__")}

    ret = {
        "args": args,
        "kwargs": kwargs,
    }
    return ret


def raw_arg(*args, **kwargs):
    """
    Output the given args and kwargs

    CLI Example:

    .. code-block:: bash

        salt-run test.arg foo __bar=baz
    """
    ret = {
        "args": args,
        "kwargs": kwargs,
    }
    return ret


def metasyntactic(locality="us"):
    """
    Return common metasyntactic variables for the given locality

    CLI Example:

    .. code-block:: bash

        salt-run test.metasyntactic locality=uk
    """
    lookup = {
        "us": [
            "foo",
            "bar",
            "baz",
            "qux",
            "quux",
            "quuz",
            "corge",
            "grault",
            "garply",
            "waldo",
            "fred",
            "plugh",
            "xyzzy",
            "thud",
        ],
        "uk": ["wibble", "wobble", "wubble", "flob"],
    }
    return lookup.get(locality, None)


def stdout_print():
    """
    Print 'foo' and return 'bar'

    CLI Example:

    .. code-block:: bash

        salt-run test.stdout_print
    """
    print("foo")
    return "bar"


def sleep(s_time=10):
    """
    Sleep t seconds, then return True

    CLI Example:

    .. code-block:: bash

        salt-run test.sleep s_time=5
    """
    print(s_time)
    time.sleep(s_time)
    return True


def stream():
    """
    Fire a stream of 100 test events, then return True

    CLI Example:

    .. code-block:: bash

        salt-run test.stream
    """
    ret = True
    for i in range(1, 100):
        __jid_event__.fire_event({"message": f"Runner is {i}% done"}, "progress")
        time.sleep(0.1)
    return ret


def get_opts():
    """
    .. versionadded:: 2018.3.0

    Return the configuration options of the master.

    CLI Example:

    .. code-block:: bash

        salt-run test.get_opts
    """
    return __opts__



"""
Azure (ARM) Compute Execution Module

.. versionadded:: 2019.2.0

.. warning::

    This cloud provider will be removed from Salt in version 3007 in favor of
    the `saltext.azurerm Salt Extension
    <https://github.com/salt-extensions/saltext-azurerm>`_

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as keyword arguments
    to every function in order to work properly.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

**cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.
    Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

"""

# Python libs

import logging
from functools import wraps

import salt.utils.azurearm

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.compute.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__virtualname__ = "azurearm_compute"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_LIBS:
        return (
            False,
            "The following dependencies are required to use the AzureARM modules: "
            "Microsoft Azure SDK for Python >= 2.0rc6, "
            "MS REST Azure (msrestazure) >= 0.4",
        )

    return __virtualname__


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Chlorine",
            "The 'azurearm' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3007 in favor of the "
            "saltext.azurerm Salt Extension. "
            "(https://github.com/salt-extensions/saltext-azurerm)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def availability_set_create_or_update(
    name, resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    Create or update an availability set.

    :param name: The availability set to create.

    :param resource_group: The resource group name assigned to the
        availability set.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.availability_set_create_or_update testset testgroup

    """
    if "location" not in kwargs:
        rg_props = __salt__["azurearm_resource.resource_group_get"](
            resource_group, **kwargs
        )

        if "error" in rg_props:
            log.error("Unable to determine location from resource group specified.")
            return False
        kwargs["location"] = rg_props["location"]

    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)

    # Use VM names to link to the IDs of existing VMs.
    if isinstance(kwargs.get("virtual_machines"), list):
        vm_list = []
        for vm_name in kwargs.get("virtual_machines"):
            vm_instance = __salt__["azurearm_compute.virtual_machine_get"](
                name=vm_name, resource_group=resource_group, **kwargs
            )
            if "error" not in vm_instance:
                vm_list.append({"id": str(vm_instance["id"])})
        kwargs["virtual_machines"] = vm_list

    try:
        setmodel = __utils__["azurearm.create_object_model"](
            "compute", "AvailabilitySet", **kwargs
        )
    except TypeError as exc:
        result = {"error": f"The object model could not be built. ({str(exc)})"}
        return result

    try:
        av_set = compconn.availability_sets.create_or_update(
            resource_group_name=resource_group,
            availability_set_name=name,
            parameters=setmodel,
        )
        result = av_set.as_dict()

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {"error": f"The object model could not be parsed. ({str(exc)})"}

    return result


@_deprecation_message
def availability_set_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Delete an availability set.

    :param name: The availability set to delete.

    :param resource_group: The resource group name assigned to the
        availability set.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.availability_set_delete testset testgroup

    """
    result = False
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        compconn.availability_sets.delete(
            resource_group_name=resource_group, availability_set_name=name
        )
        result = True

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)

    return result


@_deprecation_message
def availability_set_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Get a dictionary representing an availability set's properties.

    :param name: The availability set to get.

    :param resource_group: The resource group name assigned to the
        availability set.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.availability_set_get testset testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        av_set = compconn.availability_sets.get(
            resource_group_name=resource_group, availability_set_name=name
        )
        result = av_set.as_dict()

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def availability_sets_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all availability sets within a resource group.

    :param resource_group: The resource group name to list availability
        sets within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.availability_sets_list testgroup

    """
    result = {}
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        avail_sets = __utils__["azurearm.paged_object_to_list"](
            compconn.availability_sets.list(resource_group_name=resource_group)
        )

        for avail_set in avail_sets:
            result[avail_set["name"]] = avail_set
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def availability_sets_list_available_sizes(
    name, resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    List all available virtual machine sizes that can be used to
    to create a new virtual machine in an existing availability set.

    :param name: The availability set name to list available
        virtual machine sizes within.

    :param resource_group: The resource group name to list available
        availability set sizes within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.availability_sets_list_available_sizes testset testgroup

    """
    result = {}
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        sizes = __utils__["azurearm.paged_object_to_list"](
            compconn.availability_sets.list_available_sizes(
                resource_group_name=resource_group, availability_set_name=name
            )
        )

        for size in sizes:
            result[size["name"]] = size
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_capture(
    name, destination_name, resource_group, prefix="capture-", overwrite=False, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Captures the VM by copying virtual hard disks of the VM and outputs
    a template that can be used to create similar VMs.

    :param name: The name of the virtual machine.

    :param destination_name: The destination container name.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    :param prefix: (Default: 'capture-') The captured virtual hard disk's name prefix.

    :param overwrite: (Default: False) Overwrite the destination disk in case of conflict.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_capture testvm testcontainer testgroup

    """
    # pylint: disable=invalid-name
    VirtualMachineCaptureParameters = getattr(
        azure.mgmt.compute.models, "VirtualMachineCaptureParameters"
    )

    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.capture(
            resource_group_name=resource_group,
            vm_name=name,
            parameters=VirtualMachineCaptureParameters(
                vhd_prefix=prefix,
                destination_container_name=destination_name,
                overwrite_vhds=overwrite,
            ),
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Retrieves information about the model view or the instance view of a
    virtual machine.

    :param name: The name of the virtual machine.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_get testvm testgroup

    """
    expand = kwargs.get("expand")

    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.get(
            resource_group_name=resource_group, vm_name=name, expand=expand
        )
        result = vm.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_convert_to_managed_disks(
    name, resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    Converts virtual machine disks from blob-based to managed disks. Virtual
    machine must be stop-deallocated before invoking this operation.

    :param name: The name of the virtual machine to convert.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_convert_to_managed_disks testvm testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.convert_to_managed_disks(
            resource_group_name=resource_group, vm_name=name
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_deallocate(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Power off a virtual machine and deallocate compute resources.

    :param name: The name of the virtual machine to deallocate.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_deallocate testvm testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.deallocate(
            resource_group_name=resource_group, vm_name=name
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_generalize(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Set the state of a virtual machine to 'generalized'.

    :param name: The name of the virtual machine.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_generalize testvm testgroup

    """
    result = False
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        compconn.virtual_machines.generalize(
            resource_group_name=resource_group, vm_name=name
        )
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)

    return result


@_deprecation_message
def virtual_machines_list(resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    List all virtual machines within a resource group.

    :param resource_group: The resource group name to list virtual
        machines within.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machines_list testgroup

    """
    result = {}
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        vms = __utils__["azurearm.paged_object_to_list"](
            compconn.virtual_machines.list(resource_group_name=resource_group)
        )
        for vm in vms:  # pylint: disable=invalid-name
            result[vm["name"]] = vm
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machines_list_all(**kwargs):
    """
    .. versionadded:: 2019.2.0

    List all virtual machines within a subscription.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machines_list_all

    """
    result = {}
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        vms = __utils__["azurearm.paged_object_to_list"](
            compconn.virtual_machines.list_all()
        )
        for vm in vms:  # pylint: disable=invalid-name
            result[vm["name"]] = vm
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machines_list_available_sizes(
    name, resource_group, **kwargs
):  # pylint: disable=invalid-name
    """
    .. versionadded:: 2019.2.0

    Lists all available virtual machine sizes to which the specified virtual
    machine can be resized.

    :param name: The name of the virtual machine.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machines_list_available_sizes testvm testgroup

    """
    result = {}
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        sizes = __utils__["azurearm.paged_object_to_list"](
            compconn.virtual_machines.list_available_sizes(
                resource_group_name=resource_group, vm_name=name
            )
        )
        for size in sizes:
            result[size["name"]] = size
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_power_off(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Power off (stop) a virtual machine.

    :param name: The name of the virtual machine to stop.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_power_off testvm testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.power_off(
            resource_group_name=resource_group, vm_name=name
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_restart(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Restart a virtual machine.

    :param name: The name of the virtual machine to restart.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_restart testvm testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.restart(
            resource_group_name=resource_group, vm_name=name
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_start(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Power on (start) a virtual machine.

    :param name: The name of the virtual machine to start.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_start testvm testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.start(
            resource_group_name=resource_group, vm_name=name
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


@_deprecation_message
def virtual_machine_redeploy(name, resource_group, **kwargs):
    """
    .. versionadded:: 2019.2.0

    Redeploy a virtual machine.

    :param name: The name of the virtual machine to redeploy.

    :param resource_group: The resource group name assigned to the
        virtual machine.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_compute.virtual_machine_redeploy testvm testgroup

    """
    compconn = __utils__["azurearm.get_client"]("compute", **kwargs)
    try:
        # pylint: disable=invalid-name
        vm = compconn.virtual_machines.redeploy(
            resource_group_name=resource_group, vm_name=name
        )
        vm.wait()
        vm_result = vm.result()
        result = vm_result.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("compute", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result



import pytest

import salt.modules.win_lgpo as win_lgpo
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules():
    return {win_lgpo: {}}


def test_get_netsh_value():
    with patch.dict(win_lgpo.__context__, {"lgpo.netsh_data": {"domain": {}}}):
        win_lgpo._set_netsh_value("domain", "state", "State", "NotConfigured")
    with patch.dict(win_lgpo.__context__, {}):
        assert win_lgpo._get_netsh_value("domain", "State") == "NotConfigured"

    context = {
        "lgpo.netsh_data": {
            "domain": {
                "State": "ONContext",
                "Inbound": "NotConfigured",
                "Outbound": "NotConfigured",
                "LocalFirewallRules": "NotConfigured",
            },
        },
    }
    with patch.dict(win_lgpo.__context__, context):
        assert win_lgpo._get_netsh_value("domain", "State") == "ONContext"


def test_set_value_error():
    with pytest.raises(ValueError):
        win_lgpo._set_netsh_value("domain", "bad_section", "junk", "junk")


def test_set_value_firewall():
    mock_context = {"lgpo.netsh_data": {"domain": "junk"}}
    with patch(
        "salt.utils.win_lgpo_netsh.set_firewall_settings", MagicMock()
    ) as mock, patch.dict(win_lgpo.__context__, mock_context):
        win_lgpo._set_netsh_value(
            profile="domain",
            section="firewallpolicy",
            option="Inbound",
            value="spongebob",
        )
        mock.assert_called_once_with(
            profile="domain",
            inbound="spongebob",
            outbound=None,
            store="lgpo",
        )


def test_set_value_settings():
    mock_context = {"lgpo.netsh_data": {"domain": "junk"}}
    with patch(
        "salt.utils.win_lgpo_netsh.set_settings", MagicMock()
    ) as mock, patch.dict(win_lgpo.__context__, mock_context):
        win_lgpo._set_netsh_value(
            profile="domain",
            section="settings",
            option="spongebob",
            value="squarepants",
        )
        mock.assert_called_once_with(
            profile="domain",
            setting="spongebob",
            value="squarepants",
            store="lgpo",
        )


def test_set_value_state():
    mock_context = {"lgpo.netsh_data": {"domain": "junk"}}
    with patch("salt.utils.win_lgpo_netsh.set_state", MagicMock()) as mock, patch.dict(
        win_lgpo.__context__, mock_context
    ):
        win_lgpo._set_netsh_value(
            profile="domain",
            section="state",
            option="junk",
            value="spongebob",
        )
        mock.assert_called_once_with(
            profile="domain",
            state="spongebob",
            store="lgpo",
        )


def test_set_value_logging_filename():
    mock_context = {"lgpo.netsh_data": {"domain": "junk"}}
    with patch(
        "salt.utils.win_lgpo_netsh.set_logging_settings", MagicMock()
    ) as mock, patch.dict(win_lgpo.__context__, mock_context):
        win_lgpo._set_netsh_value(
            profile="domain",
            section="logging",
            option="FileName",
            value="Not configured",
        )
        mock.assert_called_once_with(
            profile="domain",
            setting="FileName",
            value="notconfigured",
            store="lgpo",
        )


def test_set_value_logging_log():
    mock_context = {"lgpo.netsh_data": {"domain": "junk"}}
    with patch(
        "salt.utils.win_lgpo_netsh.set_logging_settings", MagicMock()
    ) as mock, patch.dict(win_lgpo.__context__, mock_context):
        win_lgpo._set_netsh_value(
            profile="domain",
            section="logging",
            option="LogSpongebob",
            value="Junk",
        )
        mock.assert_called_once_with(
            profile="domain",
            setting="Spongebob",
            value="Junk",
            store="lgpo",
        )



"""
Module for managing SNMP service settings on Windows servers.
The Windows feature 'SNMP-Service' must be installed.
"""

import logging

import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltInvocationError

_HKEY = "HKLM"

_SNMP_KEY = r"SYSTEM\CurrentControlSet\Services\SNMP\Parameters"
_AGENT_KEY = rf"{_SNMP_KEY}\RFC1156Agent"
_COMMUNITIES_KEY = rf"{_SNMP_KEY}\ValidCommunities"

_SNMP_GPO_KEY = r"SOFTWARE\Policies\SNMP\Parameters"
_COMMUNITIES_GPO_KEY = rf"{_SNMP_GPO_KEY}\ValidCommunities"

_PERMISSION_TYPES = {
    "None": 1,
    "Notify": 2,
    "Read Only": 4,
    "Read Write": 8,
    "Read Create": 16,
}
_SERVICE_TYPES = {
    "None": 0,
    "Physical": 1,
    "Datalink and subnetwork": 2,
    "Internet": 4,
    "End-to-end": 8,
    "Applications": 64,
}

_LOG = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "win_snmp"


def __virtual__():
    """
    Only works on Windows systems.
    """
    if not salt.utils.platform.is_windows():
        return False, "Module win_snmp: Requires Windows"

    if not __utils__["reg.key_exists"](_HKEY, _SNMP_KEY):
        return False, "Module win_snmp: SNMP not installed"

    return __virtualname__


def get_agent_service_types():
    """
    Get the sysServices types that can be configured.

    Returns:
        list: A list of service types.

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.get_agent_service_types
    """
    return list(_SERVICE_TYPES)


def get_permission_types():
    """
    Get the permission types that can be configured for communities.

    Returns:
        list: A list of permission types.

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.get_permission_types
    """
    return list(_PERMISSION_TYPES)


def get_agent_settings():
    """
    Determine the value of the SNMP sysContact, sysLocation, and sysServices
    settings.

    Returns:
        dict: A dictionary of the agent settings.

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.get_agent_settings
    """
    ret = dict()
    sorted_types = sorted(_SERVICE_TYPES.items(), key=lambda x: (-x[1], x[0]))

    ret["services"] = list()
    ret["contact"] = (__utils__["reg.read_value"](_HKEY, _AGENT_KEY, "sysContact"))[
        "vdata"
    ]

    ret["location"] = (__utils__["reg.read_value"](_HKEY, _AGENT_KEY, "sysLocation"))[
        "vdata"
    ]

    current_bitmask = (__utils__["reg.read_value"](_HKEY, _AGENT_KEY, "sysServices"))[
        "vdata"
    ]

    if current_bitmask == 0:
        ret["services"].append(sorted_types[-1][0])
    else:
        # sorted_types is sorted from greatest to least bitmask.
        for service, bitmask in sorted_types:
            if current_bitmask is not None and current_bitmask > 0:
                remaining_bitmask = current_bitmask - bitmask

                if remaining_bitmask >= 0:
                    current_bitmask = remaining_bitmask
                    ret["services"].append(service)
            else:
                break

    ret["services"] = sorted(ret["services"])
    return ret


def set_agent_settings(contact=None, location=None, services=None):
    """
    Manage the SNMP sysContact, sysLocation, and sysServices settings.

    Args:

        contact (:obj:`str`, optional): The SNMP contact.

        location (:obj:`str`, optional): The SNMP location.

        services (:obj:`list`, optional):
            A list of selected services. The possible service names can be found
            via ``win_snmp.get_agent_service_types``. To disable all services
            pass a list of None, ie: ['None']

            Default is ``None``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.set_agent_settings contact='Contact Name' location='Place' services='["Physical"]'
    """
    if services is not None:
        # Filter services for unique items, and sort them for comparison
        # purposes.
        services = sorted(set(services))

        # Validate the services.
        for service in services:
            if service not in _SERVICE_TYPES:
                message = "Invalid service '{}' specified. Valid services: {}".format(
                    service, get_agent_service_types()
                )
                raise SaltInvocationError(message)

    settings = {"contact": contact, "location": location, "services": services}

    current_settings = get_agent_settings()

    if settings == current_settings:
        _LOG.debug("Agent settings already contain the provided values.")
        return True

    if contact is not None:
        if contact != current_settings["contact"]:
            __utils__["reg.set_value"](
                _HKEY, _AGENT_KEY, "sysContact", contact, "REG_SZ"
            )

    if location is not None:
        if location != current_settings["location"]:
            __utils__["reg.set_value"](
                _HKEY, _AGENT_KEY, "sysLocation", location, "REG_SZ"
            )

    if services is not None:
        if set(services) != set(current_settings["services"]):
            # Calculate the total value. Produces 0 if an empty list was provided,
            # corresponding to the None _SERVICE_TYPES value.
            vdata = sum(_SERVICE_TYPES[service] for service in services)

            _LOG.debug("Setting sysServices vdata to: %s", vdata)

            __utils__["reg.set_value"](
                _HKEY, _AGENT_KEY, "sysServices", vdata, "REG_DWORD"
            )

    # Get the fields post-change so that we can verify tht all values
    # were modified successfully. Track the ones that weren't.
    new_settings = get_agent_settings()
    failed_settings = dict()

    for setting in settings:
        if settings[setting] is not None and settings[setting] != new_settings[setting]:
            failed_settings[setting] = settings[setting]

    if failed_settings:
        _LOG.error("Unable to configure agent settings: %s", failed_settings)
        return False

    _LOG.debug("Agent settings configured successfully: %s", settings.keys())
    return True


def get_auth_traps_enabled():
    """
    Determine whether the host is configured to send authentication traps.

    Returns:
        bool: ``True`` if traps are enabled, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.get_auth_traps_enabled
    """
    reg_ret = __utils__["reg.read_value"](_HKEY, _SNMP_KEY, "EnableAuthenticationTraps")

    if reg_ret["vdata"] == "(value not set)":
        return False
    return bool(reg_ret["vdata"] or 0)


def set_auth_traps_enabled(status=True):
    """
    Manage the sending of authentication traps.

    Args:

        status (:obj:`bool`, optional):
            ``True`` to enable traps. ``False`` to disable.

            Default is ``True``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.set_auth_traps_enabled status='True'
    """
    vname = "EnableAuthenticationTraps"
    current_status = get_auth_traps_enabled()

    if bool(status) == current_status:
        _LOG.debug("%s already contains the provided value.", vname)
        return True

    vdata = int(status)
    __utils__["reg.set_value"](_HKEY, _SNMP_KEY, vname, vdata, "REG_DWORD")

    new_status = get_auth_traps_enabled()

    if status == new_status:
        _LOG.debug("Setting %s configured successfully: %s", vname, vdata)
        return True
    _LOG.error("Unable to configure %s with value: %s", vname, vdata)
    return False


def get_community_names():
    """
    Get the current accepted SNMP community names and their permissions.

    If community names are being managed by Group Policy, those values will be
    returned instead like this:

    .. code-block:: bash

        TestCommunity:
            Managed by GPO

    Community names managed normally will denote the permission instead:

    .. code-block:: bash

        TestCommunity:
            Read Only

    Returns:
        dict: A dictionary of community names and permissions.

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.get_community_names
    """
    ret = dict()

    # Look in GPO settings first
    if __utils__["reg.key_exists"](_HKEY, _COMMUNITIES_GPO_KEY):

        _LOG.debug("Loading communities from Group Policy settings")

        current_values = __utils__["reg.list_values"](_HKEY, _COMMUNITIES_GPO_KEY)

        # GPO settings are different in that they do not designate permissions
        # They are a numbered list of communities like so:
        #
        # {1: "community 1",
        #  2: "community 2"}
        #
        # Denote that it is being managed by Group Policy.
        #
        # community 1:
        #     Managed by GPO
        # community 2:
        #     Managed by GPO
        if isinstance(current_values, list):
            for current_value in current_values:

                # Ignore error values
                if not isinstance(current_value, dict):
                    continue

                ret[current_value["vdata"]] = "Managed by GPO"

    if not ret:

        _LOG.debug("Loading communities from SNMP settings")

        current_values = __utils__["reg.list_values"](_HKEY, _COMMUNITIES_KEY)

        # The communities are stored as the community name with a numeric
        # permission value. Like this (4 = Read Only):
        #
        # {"community 1": 4,
        #  "community 2": 4}
        #
        # Convert the numeric value to the text equivalent, as present in the
        # Windows SNMP service GUI.
        #
        # community 1:
        #     Read Only
        # community 2:
        #     Read Only
        if isinstance(current_values, list):
            for current_value in current_values:

                # Ignore error values
                if not isinstance(current_value, dict):
                    continue

                permissions = ""
                for permission_name in _PERMISSION_TYPES:
                    if current_value["vdata"] == _PERMISSION_TYPES[permission_name]:
                        permissions = permission_name
                        break
                ret[current_value["vname"]] = permissions

    if not ret:
        _LOG.debug("Unable to find existing communities.")
    return ret


def set_community_names(communities):
    """
    Manage the SNMP accepted community names and their permissions.

    .. note::
        Settings managed by Group Policy will always take precedence over those
        set using the SNMP interface. Therefore if this function finds Group
        Policy settings it will raise a CommandExecutionError

    Args:

        communities (dict):
            A dictionary of SNMP community names and permissions. The possible
            permissions can be found via ``win_snmp.get_permission_types``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    Raises:
        CommandExecutionError:
            If SNMP settings are being managed by Group Policy

    CLI Example:

    .. code-block:: bash

        salt '*' win_snmp.set_community_names communities='{"TestCommunity": "Read Only"}'
    """
    values = dict()

    if __utils__["reg.key_exists"](_HKEY, _COMMUNITIES_GPO_KEY):
        _LOG.debug("Communities on this system are managed by Group Policy")
        raise CommandExecutionError(
            "Communities on this system are managed by Group Policy"
        )

    current_communities = get_community_names()

    if communities == current_communities:
        _LOG.debug("Communities already contain the provided values.")
        return True

    for vname in communities:
        if not communities[vname]:
            communities[vname] = "None"
        try:
            vdata = _PERMISSION_TYPES[communities[vname]]
        except KeyError:
            raise SaltInvocationError(
                "Invalid permission '{}' specified. Valid permissions: {}".format(
                    communities[vname], _PERMISSION_TYPES.keys()
                )
            )
        values[vname] = vdata

    # Check current communities.
    for current_vname in current_communities:
        if current_vname in values:
            # Modify existing communities that have a different permission value.
            if current_communities[current_vname] != values[current_vname]:
                __utils__["reg.set_value"](
                    _HKEY,
                    _COMMUNITIES_KEY,
                    current_vname,
                    values[current_vname],
                    "REG_DWORD",
                )
        else:
            # Remove current communities that weren't provided.
            __utils__["reg.delete_value"](_HKEY, _COMMUNITIES_KEY, current_vname)

    # Create any new communities.
    for vname in values:
        if vname not in current_communities:
            __utils__["reg.set_value"](
                _HKEY, _COMMUNITIES_KEY, vname, values[vname], "REG_DWORD"
            )

    # Get the fields post-change so that we can verify tht all values
    # were modified successfully. Track the ones that weren't.
    new_communities = get_community_names()
    failed_communities = dict()

    for new_vname in new_communities:
        if new_vname not in communities:
            failed_communities[new_vname] = None

    for vname in communities:
        if communities[vname] != new_communities[vname]:
            failed_communities[vname] = communities[vname]

    if failed_communities:
        _LOG.error("Unable to configure communities: %s", failed_communities)
        return False
    _LOG.debug("Communities configured successfully: %s", communities.keys())
    return True



#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utility classes to write to and read from non-blocking files and sockets.

Contents:

* `BaseIOStream`: Generic interface for reading and writing.
* `IOStream`: Implementation of BaseIOStream using non-blocking sockets.
* `SSLIOStream`: SSL-aware version of IOStream.
* `PipeIOStream`: Pipe-based IOStream implementation.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import collections
import errno
import numbers
import os
import socket
import sys
import re

from salt.ext.tornado.concurrent import TracebackFuture
from salt.ext.tornado import ioloop
from salt.ext.tornado.log import gen_log, app_log
from salt.ext.tornado.netutil import ssl_wrap_socket, ssl_match_hostname, SSLCertificateError, _client_ssl_defaults, _server_ssl_defaults
from salt.ext.tornado import stack_context
from salt.ext.tornado.util import errno_from_exception

try:
    from salt.ext.tornado.platform.posix import _set_nonblocking
except ImportError:
    _set_nonblocking = None

try:
    import ssl
except ImportError:
    # ssl is not available on Google App Engine
    ssl = None

# These errnos indicate that a non-blocking operation must be retried
# at a later time.  On most platforms they're the same value, but on
# some they differ.
_ERRNO_WOULDBLOCK = (errno.EWOULDBLOCK, errno.EAGAIN)

if hasattr(errno, "WSAEWOULDBLOCK"):
    _ERRNO_WOULDBLOCK += (errno.WSAEWOULDBLOCK,)  # type: ignore

# These errnos indicate that a connection has been abruptly terminated.
# They should be caught and handled less noisily than other errors.
_ERRNO_CONNRESET = (errno.ECONNRESET, errno.ECONNABORTED, errno.EPIPE,
                    errno.ETIMEDOUT)

if hasattr(errno, "WSAECONNRESET"):
    _ERRNO_CONNRESET += (errno.WSAECONNRESET, errno.WSAECONNABORTED, errno.WSAETIMEDOUT)  # type: ignore

if sys.platform == 'darwin':
    # OSX appears to have a race condition that causes send(2) to return
    # EPROTOTYPE if called while a socket is being torn down:
    # http://erickt.github.io/blog/2014/11/19/adventures-in-debugging-a-potential-osx-kernel-bug/
    # Since the socket is being closed anyway, treat this as an ECONNRESET
    # instead of an unexpected error.
    _ERRNO_CONNRESET += (errno.EPROTOTYPE,)  # type: ignore

# More non-portable errnos:
_ERRNO_INPROGRESS = (errno.EINPROGRESS,)

if hasattr(errno, "WSAEINPROGRESS"):
    _ERRNO_INPROGRESS += (errno.WSAEINPROGRESS,)  # type: ignore

_WINDOWS = sys.platform.startswith('win')


class StreamClosedError(IOError):
    """Exception raised by `IOStream` methods when the stream is closed.

    Note that the close callback is scheduled to run *after* other
    callbacks on the stream (to allow for buffered data to be processed),
    so you may see this error before you see the close callback.

    The ``real_error`` attribute contains the underlying error that caused
    the stream to close (if any).

    .. versionchanged:: 4.3
       Added the ``real_error`` attribute.
    """
    def __init__(self, real_error=None):
        super(StreamClosedError, self).__init__('Stream is closed')
        self.real_error = real_error


class UnsatisfiableReadError(Exception):
    """Exception raised when a read cannot be satisfied.

    Raised by ``read_until`` and ``read_until_regex`` with a ``max_bytes``
    argument.
    """
    pass


class StreamBufferFullError(Exception):
    """Exception raised by `IOStream` methods when the buffer is full.
    """


class BaseIOStream(object):
    """A utility class to write to and read from a non-blocking file or socket.

    We support a non-blocking ``write()`` and a family of ``read_*()`` methods.
    All of the methods take an optional ``callback`` argument and return a
    `.Future` only if no callback is given.  When the operation completes,
    the callback will be run or the `.Future` will resolve with the data
    read (or ``None`` for ``write()``).  All outstanding ``Futures`` will
    resolve with a `StreamClosedError` when the stream is closed; users
    of the callback interface will be notified via
    `.BaseIOStream.set_close_callback` instead.

    When a stream is closed due to an error, the IOStream's ``error``
    attribute contains the exception object.

    Subclasses must implement `fileno`, `close_fd`, `write_to_fd`,
    `read_from_fd`, and optionally `get_fd_error`.
    """
    def __init__(self, io_loop=None, max_buffer_size=None,
                 read_chunk_size=None, max_write_buffer_size=None):
        """`BaseIOStream` constructor.

        :arg io_loop: The `.IOLoop` to use; defaults to `.IOLoop.current`.
                      Deprecated since Tornado 4.1.
        :arg max_buffer_size: Maximum amount of incoming data to buffer;
            defaults to 100MB.
        :arg read_chunk_size: Amount of data to read at one time from the
            underlying transport; defaults to 64KB.
        :arg max_write_buffer_size: Amount of outgoing data to buffer;
            defaults to unlimited.

        .. versionchanged:: 4.0
           Add the ``max_write_buffer_size`` parameter.  Changed default
           ``read_chunk_size`` to 64KB.
        """
        self.io_loop = io_loop or ioloop.IOLoop.current()
        self.max_buffer_size = max_buffer_size or 104857600
        # A chunk size that is too close to max_buffer_size can cause
        # spurious failures.
        self.read_chunk_size = min(read_chunk_size or 65536,
                                   self.max_buffer_size // 2)
        self.max_write_buffer_size = max_write_buffer_size
        self.error = None
        self._read_buffer = bytearray()
        self._read_buffer_pos = 0
        self._read_buffer_size = 0
        self._write_buffer = bytearray()
        self._write_buffer_pos = 0
        self._write_buffer_size = 0
        self._write_buffer_frozen = False
        self._total_write_index = 0
        self._total_write_done_index = 0
        self._pending_writes_while_frozen = []
        self._read_delimiter = None
        self._read_regex = None
        self._read_max_bytes = None
        self._read_bytes = None
        self._read_partial = False
        self._read_until_close = False
        self._read_callback = None
        self._read_future = None
        self._streaming_callback = None
        self._write_callback = None
        self._write_futures = collections.deque()
        self._close_callback = None
        self._connect_callback = None
        self._connect_future = None
        # _ssl_connect_future should be defined in SSLIOStream
        # but it's here so we can clean it up in maybe_run_close_callback.
        # TODO: refactor that so subclasses can add additional futures
        # to be cancelled.
        self._ssl_connect_future = None
        self._connecting = False
        self._state = None
        self._pending_callbacks = 0
        self._closed = False

    def fileno(self):
        """Returns the file descriptor for this stream."""
        raise NotImplementedError()

    def close_fd(self):
        """Closes the file underlying this stream.

        ``close_fd`` is called by `BaseIOStream` and should not be called
        elsewhere; other users should call `close` instead.
        """
        raise NotImplementedError()

    def write_to_fd(self, data):
        """Attempts to write ``data`` to the underlying file.

        Returns the number of bytes written.
        """
        raise NotImplementedError()

    def read_from_fd(self):
        """Attempts to read from the underlying file.

        Returns ``None`` if there was nothing to read (the socket
        returned `~errno.EWOULDBLOCK` or equivalent), otherwise
        returns the data.  When possible, should return no more than
        ``self.read_chunk_size`` bytes at a time.
        """
        raise NotImplementedError()

    def get_fd_error(self):
        """Returns information about any error on the underlying file.

        This method is called after the `.IOLoop` has signaled an error on the
        file descriptor, and should return an Exception (such as `socket.error`
        with additional information, or None if no such information is
        available.
        """
        return None

    def read_until_regex(self, regex, callback=None, max_bytes=None):
        """Asynchronously read until we have matched the given regex.

        The result includes the data that matches the regex and anything
        that came before it.  If a callback is given, it will be run
        with the data as an argument; if not, this method returns a
        `.Future`.

        If ``max_bytes`` is not None, the connection will be closed
        if more than ``max_bytes`` bytes have been read and the regex is
        not satisfied.

        .. versionchanged:: 4.0
            Added the ``max_bytes`` argument.  The ``callback`` argument is
            now optional and a `.Future` will be returned if it is omitted.
        """
        future = self._set_read_callback(callback)
        self._read_regex = re.compile(regex)
        self._read_max_bytes = max_bytes
        try:
            self._try_inline_read()
        except UnsatisfiableReadError as e:
            # Handle this the same way as in _handle_events.
            gen_log.info("Unsatisfiable read, closing connection: %s" % e)
            self.close(exc_info=True)
            return future
        except:
            if future is not None:
                # Ensure that the future doesn't log an error because its
                # failure was never examined.
                future.add_done_callback(lambda f: f.exception())
            raise
        return future

    def read_until(self, delimiter, callback=None, max_bytes=None):
        """Asynchronously read until we have found the given delimiter.

        The result includes all the data read including the delimiter.
        If a callback is given, it will be run with the data as an argument;
        if not, this method returns a `.Future`.

        If ``max_bytes`` is not None, the connection will be closed
        if more than ``max_bytes`` bytes have been read and the delimiter
        is not found.

        .. versionchanged:: 4.0
            Added the ``max_bytes`` argument.  The ``callback`` argument is
            now optional and a `.Future` will be returned if it is omitted.
        """
        future = self._set_read_callback(callback)
        self._read_delimiter = delimiter
        self._read_max_bytes = max_bytes
        try:
            self._try_inline_read()
        except UnsatisfiableReadError as e:
            # Handle this the same way as in _handle_events.
            gen_log.info("Unsatisfiable read, closing connection: %s" % e)
            self.close(exc_info=True)
            return future
        except:
            if future is not None:
                future.add_done_callback(lambda f: f.exception())
            raise
        return future

    def read_bytes(self, num_bytes, callback=None, streaming_callback=None,
                   partial=False):
        """Asynchronously read a number of bytes.

        If a ``streaming_callback`` is given, it will be called with chunks
        of data as they become available, and the final result will be empty.
        Otherwise, the result is all the data that was read.
        If a callback is given, it will be run with the data as an argument;
        if not, this method returns a `.Future`.

        If ``partial`` is true, the callback is run as soon as we have
        any bytes to return (but never more than ``num_bytes``)

        .. versionchanged:: 4.0
            Added the ``partial`` argument.  The callback argument is now
            optional and a `.Future` will be returned if it is omitted.
        """
        future = self._set_read_callback(callback)
        assert isinstance(num_bytes, numbers.Integral)
        self._read_bytes = num_bytes
        self._read_partial = partial
        self._streaming_callback = stack_context.wrap(streaming_callback)
        try:
            self._try_inline_read()
        except:
            if future is not None:
                future.add_done_callback(lambda f: f.exception())
            raise
        return future

    def read_until_close(self, callback=None, streaming_callback=None):
        """Asynchronously reads all data from the socket until it is closed.

        If a ``streaming_callback`` is given, it will be called with chunks
        of data as they become available, and the final result will be empty.
        Otherwise, the result is all the data that was read.
        If a callback is given, it will be run with the data as an argument;
        if not, this method returns a `.Future`.

        Note that if a ``streaming_callback`` is used, data will be
        read from the socket as quickly as it becomes available; there
        is no way to apply backpressure or cancel the reads. If flow
        control or cancellation are desired, use a loop with
        `read_bytes(partial=True) <.read_bytes>` instead.

        .. versionchanged:: 4.0
            The callback argument is now optional and a `.Future` will
            be returned if it is omitted.

        """
        future = self._set_read_callback(callback)
        self._streaming_callback = stack_context.wrap(streaming_callback)
        if self.closed():
            if self._streaming_callback is not None:
                self._run_read_callback(self._read_buffer_size, True)
            self._run_read_callback(self._read_buffer_size, False)
            return future
        self._read_until_close = True
        try:
            self._try_inline_read()
        except:
            if future is not None:
                future.add_done_callback(lambda f: f.exception())
            raise
        return future

    def write(self, data, callback=None):
        """Asynchronously write the given data to this stream.

        If ``callback`` is given, we call it when all of the buffered write
        data has been successfully written to the stream. If there was
        previously buffered write data and an old write callback, that
        callback is simply overwritten with this new callback.

        If no ``callback`` is given, this method returns a `.Future` that
        resolves (with a result of ``None``) when the write has been
        completed.

        The ``data`` argument may be of type `bytes` or `memoryview`.

        .. versionchanged:: 4.0
            Now returns a `.Future` if no callback is given.

        .. versionchanged:: 4.5
            Added support for `memoryview` arguments.
        """
        self._check_closed()
        if data:
            if (self.max_write_buffer_size is not None and
                    self._write_buffer_size + len(data) > self.max_write_buffer_size):
                raise StreamBufferFullError("Reached maximum write buffer size")
            if self._write_buffer_frozen:
                self._pending_writes_while_frozen.append(data)
            else:
                self._write_buffer += data
                self._write_buffer_size += len(data)
            self._total_write_index += len(data)
        if callback is not None:
            self._write_callback = stack_context.wrap(callback)
            future = None
        else:
            future = TracebackFuture()
            future.add_done_callback(lambda f: f.exception())
            self._write_futures.append((self._total_write_index, future))
        if not self._connecting:
            self._handle_write()
            if self._write_buffer_size:
                self._add_io_state(self.io_loop.WRITE)
            self._maybe_add_error_listener()
        return future

    def set_close_callback(self, callback):
        """Call the given callback when the stream is closed.

        This is not necessary for applications that use the `.Future`
        interface; all outstanding ``Futures`` will resolve with a
        `StreamClosedError` when the stream is closed.
        """
        self._close_callback = stack_context.wrap(callback)
        self._maybe_add_error_listener()

    def close(self, exc_info=False):
        """Close this stream.

        If ``exc_info`` is true, set the ``error`` attribute to the current
        exception from `sys.exc_info` (or if ``exc_info`` is a tuple,
        use that instead of `sys.exc_info`).
        """
        if not self.closed():
            if exc_info:
                if not isinstance(exc_info, tuple):
                    exc_info = sys.exc_info()
                if any(exc_info):
                    self.error = exc_info[1]
            if self._read_until_close:
                if (self._streaming_callback is not None and
                        self._read_buffer_size):
                    self._run_read_callback(self._read_buffer_size, True)
                self._read_until_close = False
                self._run_read_callback(self._read_buffer_size, False)
            if self._state is not None:
                self.io_loop.remove_handler(self.fileno())
                self._state = None
            self.close_fd()
            self._closed = True
        self._maybe_run_close_callback()

    def _maybe_run_close_callback(self):
        # If there are pending callbacks, don't run the close callback
        # until they're done (see _maybe_add_error_handler)
        if self.closed() and self._pending_callbacks == 0:
            futures = []
            if self._read_future is not None:
                futures.append(self._read_future)
                self._read_future = None
            futures += [future for _, future in self._write_futures]
            self._write_futures.clear()
            if self._connect_future is not None:
                futures.append(self._connect_future)
                self._connect_future = None
            if self._ssl_connect_future is not None:
                futures.append(self._ssl_connect_future)
                self._ssl_connect_future = None
            for future in futures:
                future.set_exception(StreamClosedError(real_error=self.error))
            if self._close_callback is not None:
                cb = self._close_callback
                self._close_callback = None
                self._run_callback(cb)
            # Delete any unfinished callbacks to break up reference cycles.
            self._read_callback = self._write_callback = None
            # Clear the buffers so they can be cleared immediately even
            # if the IOStream object is kept alive by a reference cycle.
            # TODO: Clear the read buffer too; it currently breaks some tests.
            self._write_buffer = None
            self._write_buffer_size = 0

    def reading(self):
        """Returns true if we are currently reading from the stream."""
        return self._read_callback is not None or self._read_future is not None

    def writing(self):
        """Returns true if we are currently writing to the stream."""
        return self._write_buffer_size > 0

    def closed(self):
        """Returns true if the stream has been closed."""
        return self._closed

    def set_nodelay(self, value):
        """Sets the no-delay flag for this stream.

        By default, data written to TCP streams may be held for a time
        to make the most efficient use of bandwidth (according to
        Nagle's algorithm).  The no-delay flag requests that data be
        written as soon as possible, even if doing so would consume
        additional bandwidth.

        This flag is currently defined only for TCP-based ``IOStreams``.

        .. versionadded:: 3.1
        """
        pass

    def _handle_events(self, fd, events):
        if self.closed():
            gen_log.warning("Got events for closed stream %s", fd)
            return
        try:
            if self._connecting:
                # Most IOLoops will report a write failed connect
                # with the WRITE event, but SelectIOLoop reports a
                # READ as well so we must check for connecting before
                # either.
                self._handle_connect()
            if self.closed():
                return
            if events & self.io_loop.READ:
                self._handle_read()
            if self.closed():
                return
            if events & self.io_loop.WRITE:
                self._handle_write()
            if self.closed():
                return
            if events & self.io_loop.ERROR:
                self.error = self.get_fd_error()
                # We may have queued up a user callback in _handle_read or
                # _handle_write, so don't close the IOStream until those
                # callbacks have had a chance to run.
                self.io_loop.add_callback(self.close)
                return
            state = self.io_loop.ERROR
            if self.reading():
                state |= self.io_loop.READ
            if self.writing():
                state |= self.io_loop.WRITE
            if state == self.io_loop.ERROR and self._read_buffer_size == 0:
                # If the connection is idle, listen for reads too so
                # we can tell if the connection is closed.  If there is
                # data in the read buffer we won't run the close callback
                # yet anyway, so we don't need to listen in this case.
                state |= self.io_loop.READ
            if state != self._state:
                assert self._state is not None, \
                    "shouldn't happen: _handle_events without self._state"
                self._state = state
                self.io_loop.update_handler(self.fileno(), self._state)
        except UnsatisfiableReadError as e:
            gen_log.info("Unsatisfiable read, closing connection: %s" % e)
            self.close(exc_info=True)
        except Exception:
            gen_log.error("Uncaught exception, closing connection.",
                          exc_info=True)
            self.close(exc_info=True)
            raise

    def _run_callback(self, callback, *args):
        def wrapper():
            self._pending_callbacks -= 1
            try:
                return callback(*args)
            except Exception:
                app_log.error("Uncaught exception, closing connection.",
                              exc_info=True)
                # Close the socket on an uncaught exception from a user callback
                # (It would eventually get closed when the socket object is
                # gc'd, but we don't want to rely on gc happening before we
                # run out of file descriptors)
                self.close(exc_info=True)
                # Re-raise the exception so that IOLoop.handle_callback_exception
                # can see it and log the error
                raise
            finally:
                self._maybe_add_error_listener()
        # We schedule callbacks to be run on the next IOLoop iteration
        # rather than running them directly for several reasons:
        # * Prevents unbounded stack growth when a callback calls an
        #   IOLoop operation that immediately runs another callback
        # * Provides a predictable execution context for e.g.
        #   non-reentrant mutexes
        # * Ensures that the try/except in wrapper() is run outside
        #   of the application's StackContexts
        with stack_context.NullContext():
            # stack_context was already captured in callback, we don't need to
            # capture it again for IOStream's wrapper.  This is especially
            # important if the callback was pre-wrapped before entry to
            # IOStream (as in HTTPConnection._header_callback), as we could
            # capture and leak the wrong context here.
            self._pending_callbacks += 1
            self.io_loop.add_callback(wrapper)

    def _read_to_buffer_loop(self):
        # This method is called from _handle_read and _try_inline_read.
        try:
            if self._read_bytes is not None:
                target_bytes = self._read_bytes
            elif self._read_max_bytes is not None:
                target_bytes = self._read_max_bytes
            elif self.reading():
                # For read_until without max_bytes, or
                # read_until_close, read as much as we can before
                # scanning for the delimiter.
                target_bytes = None
            else:
                target_bytes = 0
            next_find_pos = 0
            # Pretend to have a pending callback so that an EOF in
            # _read_to_buffer doesn't trigger an immediate close
            # callback.  At the end of this method we'll either
            # establish a real pending callback via
            # _read_from_buffer or run the close callback.
            #
            # We need two try statements here so that
            # pending_callbacks is decremented before the `except`
            # clause below (which calls `close` and does need to
            # trigger the callback)
            self._pending_callbacks += 1
            while not self.closed():
                # Read from the socket until we get EWOULDBLOCK or equivalent.
                # SSL sockets do some internal buffering, and if the data is
                # sitting in the SSL object's buffer select() and friends
                # can't see it; the only way to find out if it's there is to
                # try to read it.
                if self._read_to_buffer() == 0:
                    break

                self._run_streaming_callback()

                # If we've read all the bytes we can use, break out of
                # this loop.  We can't just call read_from_buffer here
                # because of subtle interactions with the
                # pending_callback and error_listener mechanisms.
                #
                # If we've reached target_bytes, we know we're done.
                if (target_bytes is not None and
                        self._read_buffer_size >= target_bytes):
                    break

                # Otherwise, we need to call the more expensive find_read_pos.
                # It's inefficient to do this on every read, so instead
                # do it on the first read and whenever the read buffer
                # size has doubled.
                if self._read_buffer_size >= next_find_pos:
                    pos = self._find_read_pos()
                    if pos is not None:
                        return pos
                    next_find_pos = self._read_buffer_size * 2
            return self._find_read_pos()
        finally:
            self._pending_callbacks -= 1

    def _handle_read(self):
        try:
            pos = self._read_to_buffer_loop()
        except UnsatisfiableReadError:
            raise
        except Exception as e:
            gen_log.warning("error on read: %s" % e)
            self.close(exc_info=True)
            return
        if pos is not None:
            self._read_from_buffer(pos)
            return
        else:
            self._maybe_run_close_callback()

    def _set_read_callback(self, callback):
        assert self._read_callback is None, "Already reading"
        assert self._read_future is None, "Already reading"
        if callback is not None:
            self._read_callback = stack_context.wrap(callback)
        else:
            self._read_future = TracebackFuture()
        return self._read_future

    def _run_read_callback(self, size, streaming):
        if streaming:
            callback = self._streaming_callback
        else:
            callback = self._read_callback
            self._read_callback = self._streaming_callback = None
            if self._read_future is not None:
                assert callback is None
                future = self._read_future
                self._read_future = None
                future.set_result(self._consume(size))
        if callback is not None:
            assert (self._read_future is None) or streaming
            self._run_callback(callback, self._consume(size))
        else:
            # If we scheduled a callback, we will add the error listener
            # afterwards.  If we didn't, we have to do it now.
            self._maybe_add_error_listener()

    def _try_inline_read(self):
        """Attempt to complete the current read operation from buffered data.

        If the read can be completed without blocking, schedules the
        read callback on the next IOLoop iteration; otherwise starts
        listening for reads on the socket.
        """
        # See if we've already got the data from a previous read
        self._run_streaming_callback()
        pos = self._find_read_pos()
        if pos is not None:
            self._read_from_buffer(pos)
            return
        self._check_closed()
        try:
            pos = self._read_to_buffer_loop()
        except Exception:
            # If there was an in _read_to_buffer, we called close() already,
            # but couldn't run the close callback because of _pending_callbacks.
            # Before we escape from this function, run the close callback if
            # applicable.
            self._maybe_run_close_callback()
            raise
        if pos is not None:
            self._read_from_buffer(pos)
            return
        # We couldn't satisfy the read inline, so either close the stream
        # or listen for new data.
        if self.closed():
            self._maybe_run_close_callback()
        else:
            self._add_io_state(ioloop.IOLoop.READ)

    def _read_to_buffer(self):
        """Reads from the socket and appends the result to the read buffer.

        Returns the number of bytes read.  Returns 0 if there is nothing
        to read (i.e. the read returns EWOULDBLOCK or equivalent).  On
        error closes the socket and raises an exception.
        """
        while True:
            try:
                chunk = self.read_from_fd()
            except (socket.error, IOError, OSError) as e:
                if errno_from_exception(e) == errno.EINTR:
                    continue
                # ssl.SSLError is a subclass of socket.error
                if self._is_connreset(e):
                    # Treat ECONNRESET as a connection close rather than
                    # an error to minimize log spam  (the exception will
                    # be available on self.error for apps that care).
                    self.close(exc_info=True)
                    return
                self.close(exc_info=True)
                raise
            break
        if chunk is None:
            return 0
        self._read_buffer += chunk
        self._read_buffer_size += len(chunk)
        if self._read_buffer_size > self.max_buffer_size:
            gen_log.error("Reached maximum read buffer size")
            self.close()
            raise StreamBufferFullError("Reached maximum read buffer size")
        return len(chunk)

    def _run_streaming_callback(self):
        if self._streaming_callback is not None and self._read_buffer_size:
            bytes_to_consume = self._read_buffer_size
            if self._read_bytes is not None:
                bytes_to_consume = min(self._read_bytes, bytes_to_consume)
                self._read_bytes -= bytes_to_consume
            self._run_read_callback(bytes_to_consume, True)

    def _read_from_buffer(self, pos):
        """Attempts to complete the currently-pending read from the buffer.

        The argument is either a position in the read buffer or None,
        as returned by _find_read_pos.
        """
        self._read_bytes = self._read_delimiter = self._read_regex = None
        self._read_partial = False
        self._run_read_callback(pos, False)

    def _find_read_pos(self):
        """Attempts to find a position in the read buffer that satisfies
        the currently-pending read.

        Returns a position in the buffer if the current read can be satisfied,
        or None if it cannot.
        """
        if (self._read_bytes is not None and
            (self._read_buffer_size >= self._read_bytes or
             (self._read_partial and self._read_buffer_size > 0))):
            num_bytes = min(self._read_bytes, self._read_buffer_size)
            return num_bytes
        elif self._read_delimiter is not None:
            # Multi-byte delimiters (e.g. '\r\n') may straddle two
            # chunks in the read buffer, so we can't easily find them
            # without collapsing the buffer.  However, since protocols
            # using delimited reads (as opposed to reads of a known
            # length) tend to be "line" oriented, the delimiter is likely
            # to be in the first few chunks.  Merge the buffer gradually
            # since large merges are relatively expensive and get undone in
            # _consume().
            if self._read_buffer:
                loc = self._read_buffer.find(self._read_delimiter,
                                             self._read_buffer_pos)
                if loc != -1:
                    loc -= self._read_buffer_pos
                    delimiter_len = len(self._read_delimiter)
                    self._check_max_bytes(self._read_delimiter,
                                          loc + delimiter_len)
                    return loc + delimiter_len
                self._check_max_bytes(self._read_delimiter,
                                      self._read_buffer_size)
        elif self._read_regex is not None:
            if self._read_buffer:
                m = self._read_regex.search(self._read_buffer,
                                            self._read_buffer_pos)
                if m is not None:
                    loc = m.end() - self._read_buffer_pos
                    self._check_max_bytes(self._read_regex, loc)
                    return loc
                self._check_max_bytes(self._read_regex, self._read_buffer_size)
        return None

    def _check_max_bytes(self, delimiter, size):
        if (self._read_max_bytes is not None and
                size > self._read_max_bytes):
            raise UnsatisfiableReadError(
                "delimiter %r not found within %d bytes" % (
                    delimiter, self._read_max_bytes))

    def _freeze_write_buffer(self, size):
        self._write_buffer_frozen = size

    def _unfreeze_write_buffer(self):
        self._write_buffer_frozen = False
        self._write_buffer += b''.join(self._pending_writes_while_frozen)
        self._write_buffer_size += sum(map(len, self._pending_writes_while_frozen))
        self._pending_writes_while_frozen[:] = []

    def _got_empty_write(self, size):
        """
        Called when a non-blocking write() failed writing anything.
        Can be overridden in subclasses.
        """

    def _handle_write(self):
        while self._write_buffer_size:
            assert self._write_buffer_size >= 0
            try:
                start = self._write_buffer_pos
                if self._write_buffer_frozen:
                    size = self._write_buffer_frozen
                elif _WINDOWS:
                    # On windows, socket.send blows up if given a
                    # write buffer that's too large, instead of just
                    # returning the number of bytes it was able to
                    # process.  Therefore we must not call socket.send
                    # with more than 128KB at a time.
                    size = 128 * 1024
                else:
                    size = self._write_buffer_size
                num_bytes = self.write_to_fd(
                    memoryview(self._write_buffer)[start:start + size])
                if num_bytes == 0:
                    self._got_empty_write(size)
                    break
                self._write_buffer_pos += num_bytes
                self._write_buffer_size -= num_bytes
                # Amortized O(1) shrink
                # (this heuristic is implemented natively in Python 3.4+
                #  but is replicated here for Python 2)
                if self._write_buffer_pos > self._write_buffer_size:
                    del self._write_buffer[:self._write_buffer_pos]
                    self._write_buffer_pos = 0
                if self._write_buffer_frozen:
                    self._unfreeze_write_buffer()
                self._total_write_done_index += num_bytes
            except (socket.error, IOError, OSError) as e:
                if e.args[0] in _ERRNO_WOULDBLOCK:
                    self._got_empty_write(size)
                    break
                else:
                    if not self._is_connreset(e):
                        # Broken pipe errors are usually caused by connection
                        # reset, and its better to not log EPIPE errors to
                        # minimize log spam
                        gen_log.warning("Write error on %s: %s",
                                        self.fileno(), e)
                    self.close(exc_info=True)
                    return

        while self._write_futures:
            index, future = self._write_futures[0]
            if index > self._total_write_done_index:
                break
            self._write_futures.popleft()
            future.set_result(None)

        if not self._write_buffer_size:
            if self._write_callback:
                callback = self._write_callback
                self._write_callback = None
                self._run_callback(callback)

    def _consume(self, loc):
        # Consume loc bytes from the read buffer and return them
        if loc == 0:
            return b""
        assert loc <= self._read_buffer_size
        # Slice the bytearray buffer into bytes, without intermediate copying
        b = (memoryview(self._read_buffer)
             [self._read_buffer_pos:self._read_buffer_pos + loc]
             ).tobytes()
        self._read_buffer_pos += loc
        self._read_buffer_size -= loc
        # Amortized O(1) shrink
        # (this heuristic is implemented natively in Python 3.4+
        #  but is replicated here for Python 2)
        if self._read_buffer_pos > self._read_buffer_size:
            del self._read_buffer[:self._read_buffer_pos]
            self._read_buffer_pos = 0
        return b

    def _check_closed(self):
        if self.closed():
            raise StreamClosedError(real_error=self.error)

    def _maybe_add_error_listener(self):
        # This method is part of an optimization: to detect a connection that
        # is closed when we're not actively reading or writing, we must listen
        # for read events.  However, it is inefficient to do this when the
        # connection is first established because we are going to read or write
        # immediately anyway.  Instead, we insert checks at various times to
        # see if the connection is idle and add the read listener then.
        if self._pending_callbacks != 0:
            return
        if self._state is None or self._state == ioloop.IOLoop.ERROR:
            if self.closed():
                self._maybe_run_close_callback()
            elif (self._read_buffer_size == 0 and
                  self._close_callback is not None):
                self._add_io_state(ioloop.IOLoop.READ)

    def _add_io_state(self, state):
        """Adds `state` (IOLoop.{READ,WRITE} flags) to our event handler.

        Implementation notes: Reads and writes have a fast path and a
        slow path.  The fast path reads synchronously from socket
        buffers, while the slow path uses `_add_io_state` to schedule
        an IOLoop callback.  Note that in both cases, the callback is
        run asynchronously with `_run_callback`.

        To detect closed connections, we must have called
        `_add_io_state` at some point, but we want to delay this as
        much as possible so we don't have to set an `IOLoop.ERROR`
        listener that will be overwritten by the next slow-path
        operation.  As long as there are callbacks scheduled for
        fast-path ops, those callbacks may do more reads.
        If a sequence of fast-path ops do not end in a slow-path op,
        (e.g. for an @asynchronous long-poll request), we must add
        the error handler.  This is done in `_run_callback` and `write`
        (since the write callback is optional so we can have a
        fast-path write with no `_run_callback`)
        """
        if self.closed():
            # connection has been closed, so there can be no future events
            return
        if self._state is None:
            self._state = ioloop.IOLoop.ERROR | state
            with stack_context.NullContext():
                self.io_loop.add_handler(
                    self.fileno(), self._handle_events, self._state)
        elif not self._state & state:
            self._state = self._state | state
            self.io_loop.update_handler(self.fileno(), self._state)

    def _is_connreset(self, exc):
        """Return true if exc is ECONNRESET or equivalent.

        May be overridden in subclasses.
        """
        return (isinstance(exc, (socket.error, IOError)) and
                errno_from_exception(exc) in _ERRNO_CONNRESET)


class IOStream(BaseIOStream):
    r"""Socket-based `IOStream` implementation.

    This class supports the read and write methods from `BaseIOStream`
    plus a `connect` method.

    The ``socket`` parameter may either be connected or unconnected.
    For server operations the socket is the result of calling
    `socket.accept <socket.socket.accept>`.  For client operations the
    socket is created with `socket.socket`, and may either be
    connected before passing it to the `IOStream` or connected with
    `IOStream.connect`.

    A very simple (and broken) HTTP client using this class:

    .. testcode::

        import tornado.ioloop
        import tornado.iostream
        import socket

        def send_request():
            stream.write(b"GET / HTTP/1.0\r\nHost: friendfeed.com\r\n\r\n")
            stream.read_until(b"\r\n\r\n", on_headers)

        def on_headers(data):
            headers = {}
            for line in data.split(b"\r\n"):
               parts = line.split(b":")
               if len(parts) == 2:
                   headers[parts[0].strip()] = parts[1].strip()
            stream.read_bytes(int(headers[b"Content-Length"]), on_body)

        def on_body(data):
            print(data)
            stream.close()
            tornado.ioloop.IOLoop.current().stop()

        if __name__ == '__main__':
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            stream = tornado.iostream.IOStream(s)
            stream.connect(("friendfeed.com", 80), send_request)
            tornado.ioloop.IOLoop.current().start()

    .. testoutput::
       :hide:

    """
    def __init__(self, socket, *args, **kwargs):
        self.socket = socket
        self.socket.setblocking(False)
        super(IOStream, self).__init__(*args, **kwargs)

    def fileno(self):
        return self.socket

    def close_fd(self):
        self.socket.close()
        self.socket = None

    def get_fd_error(self):
        errno = self.socket.getsockopt(socket.SOL_SOCKET,
                                       socket.SO_ERROR)
        return socket.error(errno, os.strerror(errno))

    def read_from_fd(self):
        try:
            chunk = self.socket.recv(self.read_chunk_size)
        except socket.error as e:
            if e.args[0] in _ERRNO_WOULDBLOCK:
                return None
            else:
                raise
        if not chunk:
            self.close()
            return None
        return chunk

    def write_to_fd(self, data):
        try:
            return self.socket.send(data)
        finally:
            # Avoid keeping to data, which can be a memoryview.
            # See https://github.com/tornadoweb/tornado/pull/2008
            del data

    def connect(self, address, callback=None, server_hostname=None):
        """Connects the socket to a remote address without blocking.

        May only be called if the socket passed to the constructor was
        not previously connected.  The address parameter is in the
        same format as for `socket.connect <socket.socket.connect>` for
        the type of socket passed to the IOStream constructor,
        e.g. an ``(ip, port)`` tuple.  Hostnames are accepted here,
        but will be resolved synchronously and block the IOLoop.
        If you have a hostname instead of an IP address, the `.TCPClient`
        class is recommended instead of calling this method directly.
        `.TCPClient` will do asynchronous DNS resolution and handle
        both IPv4 and IPv6.

        If ``callback`` is specified, it will be called with no
        arguments when the connection is completed; if not this method
        returns a `.Future` (whose result after a successful
        connection will be the stream itself).

        In SSL mode, the ``server_hostname`` parameter will be used
        for certificate validation (unless disabled in the
        ``ssl_options``) and SNI (if supported; requires Python
        2.7.9+).

        Note that it is safe to call `IOStream.write
        <BaseIOStream.write>` while the connection is pending, in
        which case the data will be written as soon as the connection
        is ready.  Calling `IOStream` read methods before the socket is
        connected works on some platforms but is non-portable.

        .. versionchanged:: 4.0
            If no callback is given, returns a `.Future`.

        .. versionchanged:: 4.2
           SSL certificates are validated by default; pass
           ``ssl_options=dict(cert_reqs=ssl.CERT_NONE)`` or a
           suitably-configured `ssl.SSLContext` to the
           `SSLIOStream` constructor to disable.
        """
        self._connecting = True
        if callback is not None:
            self._connect_callback = stack_context.wrap(callback)
            future = None
        else:
            future = self._connect_future = TracebackFuture()
        try:
            self.socket.connect(address)
        except socket.error as e:
            # In non-blocking mode we expect connect() to raise an
            # exception with EINPROGRESS or EWOULDBLOCK.
            #
            # On freebsd, other errors such as ECONNREFUSED may be
            # returned immediately when attempting to connect to
            # localhost, so handle them the same way as an error
            # reported later in _handle_connect.
            if (errno_from_exception(e) not in _ERRNO_INPROGRESS and
                    errno_from_exception(e) not in _ERRNO_WOULDBLOCK):
                if future is None:
                    gen_log.warning("Connect error on fd %s: %s",
                                    self.socket.fileno(), e)
                self.close(exc_info=True)
                return future
        self._add_io_state(self.io_loop.WRITE)
        return future

    def start_tls(self, server_side, ssl_options=None, server_hostname=None):
        """Convert this `IOStream` to an `SSLIOStream`.

        This enables protocols that begin in clear-text mode and
        switch to SSL after some initial negotiation (such as the
        ``STARTTLS`` extension to SMTP and IMAP).

        This method cannot be used if there are outstanding reads
        or writes on the stream, or if there is any data in the
        IOStream's buffer (data in the operating system's socket
        buffer is allowed).  This means it must generally be used
        immediately after reading or writing the last clear-text
        data.  It can also be used immediately after connecting,
        before any reads or writes.

        The ``ssl_options`` argument may be either an `ssl.SSLContext`
        object or a dictionary of keyword arguments for the
        `ssl.wrap_socket` function.  The ``server_hostname`` argument
        will be used for certificate validation unless disabled
        in the ``ssl_options``.

        This method returns a `.Future` whose result is the new
        `SSLIOStream`.  After this method has been called,
        any other operation on the original stream is undefined.

        If a close callback is defined on this stream, it will be
        transferred to the new stream.

        .. versionadded:: 4.0

        .. versionchanged:: 4.2
           SSL certificates are validated by default; pass
           ``ssl_options=dict(cert_reqs=ssl.CERT_NONE)`` or a
           suitably-configured `ssl.SSLContext` to disable.
        """
        if (self._read_callback or self._read_future or
                self._write_callback or self._write_futures or
                self._connect_callback or self._connect_future or
                self._pending_callbacks or self._closed or
                self._read_buffer or self._write_buffer):
            raise ValueError("IOStream is not idle; cannot convert to SSL")
        if ssl_options is None:
            if server_side:
                ssl_options = _server_ssl_defaults
            else:
                ssl_options = _client_ssl_defaults

        socket = self.socket
        self.io_loop.remove_handler(socket)
        self.socket = None
        socket = ssl_wrap_socket(socket, ssl_options,
                                 server_hostname=server_hostname,
                                 server_side=server_side,
                                 do_handshake_on_connect=False)
        orig_close_callback = self._close_callback
        self._close_callback = None

        future = TracebackFuture()
        ssl_stream = SSLIOStream(socket, ssl_options=ssl_options,
                                 io_loop=self.io_loop)
        # Wrap the original close callback so we can fail our Future as well.
        # If we had an "unwrap" counterpart to this method we would need
        # to restore the original callback after our Future resolves
        # so that repeated wrap/unwrap calls don't build up layers.

        def close_callback():
            if not future.done():
                # Note that unlike most Futures returned by IOStream,
                # this one passes the underlying error through directly
                # instead of wrapping everything in a StreamClosedError
                # with a real_error attribute. This is because once the
                # connection is established it's more helpful to raise
                # the SSLError directly than to hide it behind a
                # StreamClosedError (and the client is expecting SSL
                # issues rather than network issues since this method is
                # named start_tls).
                future.set_exception(ssl_stream.error or StreamClosedError())
            if orig_close_callback is not None:
                orig_close_callback()
        ssl_stream.set_close_callback(close_callback)
        ssl_stream._ssl_connect_callback = lambda: future.set_result(ssl_stream)
        ssl_stream.max_buffer_size = self.max_buffer_size
        ssl_stream.read_chunk_size = self.read_chunk_size
        return future

    def _handle_connect(self):
        err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err != 0:
            self.error = socket.error(err, os.strerror(err))
            # IOLoop implementations may vary: some of them return
            # an error state before the socket becomes writable, so
            # in that case a connection failure would be handled by the
            # error path in _handle_events instead of here.
            if self._connect_future is None:
                gen_log.warning("Connect error on fd %s: %s",
                                self.socket.fileno(), errno.errorcode[err])
            self.close()
            return
        if self._connect_callback is not None:
            callback = self._connect_callback
            self._connect_callback = None
            self._run_callback(callback)
        if self._connect_future is not None:
            future = self._connect_future
            self._connect_future = None
            future.set_result(self)
        self._connecting = False

    def set_nodelay(self, value):
        if (self.socket is not None and
                self.socket.family in (socket.AF_INET, socket.AF_INET6)):
            try:
                self.socket.setsockopt(socket.IPPROTO_TCP,
                                       socket.TCP_NODELAY, 1 if value else 0)
            except socket.error as e:
                # Sometimes setsockopt will fail if the socket is closed
                # at the wrong time.  This can happen with HTTPServer
                # resetting the value to false between requests.
                if e.errno != errno.EINVAL and not self._is_connreset(e):
                    raise


class SSLIOStream(IOStream):
    """A utility class to write to and read from a non-blocking SSL socket.

    If the socket passed to the constructor is already connected,
    it should be wrapped with::

        ssl.wrap_socket(sock, do_handshake_on_connect=False, **kwargs)

    before constructing the `SSLIOStream`.  Unconnected sockets will be
    wrapped when `IOStream.connect` is finished.
    """
    def __init__(self, *args, **kwargs):
        """The ``ssl_options`` keyword argument may either be an
        `ssl.SSLContext` object or a dictionary of keywords arguments
        for `ssl.wrap_socket`
        """
        self._ssl_options = kwargs.pop('ssl_options', _client_ssl_defaults)
        super(SSLIOStream, self).__init__(*args, **kwargs)
        self._ssl_accepting = True
        self._handshake_reading = False
        self._handshake_writing = False
        self._ssl_connect_callback = None
        self._server_hostname = None

        # If the socket is already connected, attempt to start the handshake.
        try:
            self.socket.getpeername()
        except socket.error:
            pass
        else:
            # Indirectly start the handshake, which will run on the next
            # IOLoop iteration and then the real IO state will be set in
            # _handle_events.
            self._add_io_state(self.io_loop.WRITE)

    def reading(self):
        return self._handshake_reading or super(SSLIOStream, self).reading()

    def writing(self):
        return self._handshake_writing or super(SSLIOStream, self).writing()

    def _got_empty_write(self, size):
        # With OpenSSL, if we couldn't write the entire buffer,
        # the very same string object must be used on the
        # next call to send.  Therefore we suppress
        # merging the write buffer after an incomplete send.
        # A cleaner solution would be to set
        # SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER, but this is
        # not yet accessible from python
        # (http://bugs.python.org/issue8240)
        self._freeze_write_buffer(size)

    def _do_ssl_handshake(self):
        # Based on code from test_ssl.py in the python stdlib
        try:
            self._handshake_reading = False
            self._handshake_writing = False
            self.socket.do_handshake()
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_WANT_READ:
                self._handshake_reading = True
                return
            elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                self._handshake_writing = True
                return
            elif err.args[0] in (ssl.SSL_ERROR_EOF,
                                 ssl.SSL_ERROR_ZERO_RETURN):
                return self.close(exc_info=True)
            elif err.args[0] == ssl.SSL_ERROR_SSL:
                try:
                    peer = self.socket.getpeername()
                except Exception:
                    peer = '(not connected)'
                gen_log.warning("SSL Error on %s %s: %s",
                                self.socket.fileno(), peer, err)
                return self.close(exc_info=True)
            raise
        except socket.error as err:
            # Some port scans (e.g. nmap in -sT mode) have been known
            # to cause do_handshake to raise EBADF and ENOTCONN, so make
            # those errors quiet as well.
            # https://groups.google.com/forum/?fromgroups#!topic/python-tornado/ApucKJat1_0
            if (self._is_connreset(err) or
                    err.args[0] in (errno.EBADF, errno.ENOTCONN)):
                return self.close(exc_info=True)
            raise
        except AttributeError:
            # On Linux, if the connection was reset before the call to
            # wrap_socket, do_handshake will fail with an
            # AttributeError.
            return self.close(exc_info=True)
        else:
            self._ssl_accepting = False
            if not self._verify_cert(self.socket.getpeercert()):
                self.close()
                return
            self._run_ssl_connect_callback()

    def _run_ssl_connect_callback(self):
        if self._ssl_connect_callback is not None:
            callback = self._ssl_connect_callback
            self._ssl_connect_callback = None
            self._run_callback(callback)
        if self._ssl_connect_future is not None:
            future = self._ssl_connect_future
            self._ssl_connect_future = None
            future.set_result(self)

    def _verify_cert(self, peercert):
        """Returns True if peercert is valid according to the configured
        validation mode and hostname.

        The ssl handshake already tested the certificate for a valid
        CA signature; the only thing that remains is to check
        the hostname.
        """
        if isinstance(self._ssl_options, dict):
            verify_mode = self._ssl_options.get('cert_reqs', ssl.CERT_NONE)
        elif isinstance(self._ssl_options, ssl.SSLContext):
            verify_mode = self._ssl_options.verify_mode
        assert verify_mode in (ssl.CERT_NONE, ssl.CERT_REQUIRED, ssl.CERT_OPTIONAL)
        if verify_mode == ssl.CERT_NONE or self._server_hostname is None:
            return True
        cert = self.socket.getpeercert()
        if cert is None and verify_mode == ssl.CERT_REQUIRED:
            gen_log.warning("No SSL certificate given")
            return False
        try:
            ssl_match_hostname(peercert, self._server_hostname)
        except SSLCertificateError as e:
            gen_log.warning("Invalid SSL certificate: %s" % e)
            return False
        else:
            return True

    def _handle_read(self):
        if self._ssl_accepting:
            self._do_ssl_handshake()
            return
        super(SSLIOStream, self)._handle_read()

    def _handle_write(self):
        if self._ssl_accepting:
            self._do_ssl_handshake()
            return
        super(SSLIOStream, self)._handle_write()

    def connect(self, address, callback=None, server_hostname=None):
        self._server_hostname = server_hostname
        # Pass a dummy callback to super.connect(), which is slightly
        # more efficient than letting it return a Future we ignore.
        super(SSLIOStream, self).connect(address, callback=lambda: None)
        return self.wait_for_handshake(callback)

    def _handle_connect(self):
        # Call the superclass method to check for errors.
        super(SSLIOStream, self)._handle_connect()
        if self.closed():
            return
        # When the connection is complete, wrap the socket for SSL
        # traffic.  Note that we do this by overriding _handle_connect
        # instead of by passing a callback to super().connect because
        # user callbacks are enqueued asynchronously on the IOLoop,
        # but since _handle_events calls _handle_connect immediately
        # followed by _handle_write we need this to be synchronous.
        #
        # The IOLoop will get confused if we swap out self.socket while the
        # fd is registered, so remove it now and re-register after
        # wrap_socket().
        self.io_loop.remove_handler(self.socket)
        old_state = self._state
        self._state = None
        self.socket = ssl_wrap_socket(self.socket, self._ssl_options,
                                      server_hostname=self._server_hostname,
                                      do_handshake_on_connect=False)
        self._add_io_state(old_state)

    def wait_for_handshake(self, callback=None):
        """Wait for the initial SSL handshake to complete.

        If a ``callback`` is given, it will be called with no
        arguments once the handshake is complete; otherwise this
        method returns a `.Future` which will resolve to the
        stream itself after the handshake is complete.

        Once the handshake is complete, information such as
        the peer's certificate and NPN/ALPN selections may be
        accessed on ``self.socket``.

        This method is intended for use on server-side streams
        or after using `IOStream.start_tls`; it should not be used
        with `IOStream.connect` (which already waits for the
        handshake to complete). It may only be called once per stream.

        .. versionadded:: 4.2
        """
        if (self._ssl_connect_callback is not None or
                self._ssl_connect_future is not None):
            raise RuntimeError("Already waiting")
        if callback is not None:
            self._ssl_connect_callback = stack_context.wrap(callback)
            future = None
        else:
            future = self._ssl_connect_future = TracebackFuture()
        if not self._ssl_accepting:
            self._run_ssl_connect_callback()
        return future

    def write_to_fd(self, data):
        try:
            return self.socket.send(data)
        except ssl.SSLError as e:
            if e.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                # In Python 3.5+, SSLSocket.send raises a WANT_WRITE error if
                # the socket is not writeable; we need to transform this into
                # an EWOULDBLOCK socket.error or a zero return value,
                # either of which will be recognized by the caller of this
                # method. Prior to Python 3.5, an unwriteable socket would
                # simply return 0 bytes written.
                return 0
            raise
        finally:
            # Avoid keeping to data, which can be a memoryview.
            # See https://github.com/tornadoweb/tornado/pull/2008
            del data

    def read_from_fd(self):
        if self._ssl_accepting:
            # If the handshake hasn't finished yet, there can't be anything
            # to read (attempting to read may or may not raise an exception
            # depending on the SSL version)
            return None
        try:
            # SSLSocket objects have both a read() and recv() method,
            # while regular sockets only have recv().
            # The recv() method blocks (at least in python 2.6) if it is
            # called when there is nothing to read, so we have to use
            # read() instead.
            chunk = self.socket.read(self.read_chunk_size)
        except ssl.SSLError as e:
            # SSLError is a subclass of socket.error, so this except
            # block must come first.
            if e.args[0] == ssl.SSL_ERROR_WANT_READ:
                return None
            else:
                raise
        except socket.error as e:
            if e.args[0] in _ERRNO_WOULDBLOCK:
                return None
            else:
                raise
        if not chunk:
            self.close()
            return None
        return chunk

    def _is_connreset(self, e):
        if isinstance(e, ssl.SSLError) and e.args[0] == ssl.SSL_ERROR_EOF:
            return True
        return super(SSLIOStream, self)._is_connreset(e)


class PipeIOStream(BaseIOStream):
    """Pipe-based `IOStream` implementation.

    The constructor takes an integer file descriptor (such as one returned
    by `os.pipe`) rather than an open file object.  Pipes are generally
    one-way, so a `PipeIOStream` can be used for reading or writing but not
    both.
    """
    def __init__(self, fd, *args, **kwargs):
        self.fd = fd
        _set_nonblocking(fd)
        super(PipeIOStream, self).__init__(*args, **kwargs)

    def fileno(self):
        return self.fd

    def close_fd(self):
        os.close(self.fd)

    def write_to_fd(self, data):
        try:
            return os.write(self.fd, data)
        finally:
            # Avoid keeping to data, which can be a memoryview.
            # See https://github.com/tornadoweb/tornado/pull/2008
            del data

    def read_from_fd(self):
        try:
            chunk = os.read(self.fd, self.read_chunk_size)
        except (IOError, OSError) as e:
            if errno_from_exception(e) in _ERRNO_WOULDBLOCK:
                return None
            elif errno_from_exception(e) == errno.EBADF:
                # If the writing half of a pipe is closed, select will
                # report it as readable but reads will fail with EBADF.
                self.close(exc_info=True)
                return None
            else:
                raise
        if not chunk:
            self.close()
            return None
        return chunk


def doctests():
    import doctest
    return doctest.DocTestSuite()



"""
unit tests for the pillar runner
"""

import logging
import os
import shutil

import pytest

import salt.runners.pillar as pillar_runner
import salt.utils.files
import salt.utils.gitfs
import salt.utils.msgpack
from tests.support.mock import MagicMock, mock_open, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        pillar_runner: {
            "__opts__": {
                "pillar_cache": True,
                "pillar_cache_backend": "disk",
                "pillar_cache_ttl": 30,
            }
        }
    }


@pytest.fixture(scope="module")
def cachedir_tree(tmp_path_factory):
    _cachedir_tree = tmp_path_factory.mktemp("cachedir")
    try:
        yield _cachedir_tree
    finally:
        shutil.rmtree(str(_cachedir_tree), ignore_errors=True)


@pytest.fixture(scope="module")
def pillar_cache_dir(cachedir_tree):
    pillar_cache_dir = cachedir_tree / "pillar_cache"
    pillar_cache_dir.mkdir()
    return pillar_cache_dir


@pytest.fixture(scope="function")
def pillar_cache_files(pillar_cache_dir):
    MINION_ID = "test-host"
    cache = {
        "CacheDisk_data": {
            MINION_ID: {
                None: {
                    "this": "one",
                    "that": "two",
                    "those": ["three", "four", "five"],
                }
            }
        },
        "CacheDisk_cachetime": {MINION_ID: 1612302460.146923},
    }
    packer = salt.utils.msgpack.Packer()
    cache_contents = packer.pack(cache)

    with salt.utils.files.fopen(
        os.path.join(str(pillar_cache_dir), MINION_ID), "wb+"
    ) as fp:
        fp.write(cache_contents)

    MINION_ID = "another-host"
    cache = {
        "CacheDisk_data": {
            MINION_ID: {
                None: {
                    "this": "six",
                    "that": "seven",
                    "those": ["eight", "nine", "ten"],
                }
            }
        },
        "CacheDisk_cachetime": {MINION_ID: 1612302460.146923},
    }
    packer = salt.utils.msgpack.Packer()
    cache_contents = packer.pack(cache)

    with salt.utils.files.fopen(
        os.path.join(str(pillar_cache_dir), MINION_ID), "wb+"
    ) as fp:
        fp.write(cache_contents)


def test_clear_pillar_cache(cachedir_tree, pillar_cache_dir, pillar_cache_files):
    """
    test pillar.clear_pillar_cache
    """

    MINION_IDS = [
        ["test-host", "another-host"],
        ["test-host"],
        ["test-host", "another-host"],
        ["test-host", "another-host"],
    ]
    _CHECK_MINIONS_RETURN = []
    for entry in MINION_IDS:
        _CHECK_MINIONS_RETURN.append({"minions": entry, "missing": []})

    with patch.dict(pillar_runner.__opts__, {"cachedir": str(cachedir_tree)}):
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(side_effect=_CHECK_MINIONS_RETURN),
        ):
            expected = {
                "test-host": {
                    "those": ["three", "four", "five"],
                    "that": "two",
                    "this": "one",
                },
                "another-host": {
                    "those": ["eight", "nine", "ten"],
                    "that": "seven",
                    "this": "six",
                },
            }
            ret = pillar_runner.show_pillar_cache()
            assert ret == expected

            ret = pillar_runner.clear_pillar_cache("test-host")
            assert ret == {}

            expected = {
                "another-host": {
                    "those": ["eight", "nine", "ten"],
                    "that": "seven",
                    "this": "six",
                }
            }
            ret = pillar_runner.show_pillar_cache()
            assert ret == expected

            ret = pillar_runner.clear_pillar_cache()
            assert ret == {}


def test_show_pillar_cache(cachedir_tree, pillar_cache_dir, pillar_cache_files):
    """
    test pillar.clear_pillar_cache
    """

    MINION_IDS = [["test-host", "another-host"], ["test-host"]]

    _CHECK_MINIONS_RETURN = []
    for entry in MINION_IDS:
        _CHECK_MINIONS_RETURN.append({"minions": entry, "missing": []})

    with patch.dict(pillar_runner.__opts__, {"cachedir": str(cachedir_tree)}):
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(side_effect=_CHECK_MINIONS_RETURN),
        ):
            expected = {
                "test-host": {
                    "those": ["three", "four", "five"],
                    "that": "two",
                    "this": "one",
                },
                "another-host": {
                    "those": ["eight", "nine", "ten"],
                    "that": "seven",
                    "this": "six",
                },
            }
            ret = pillar_runner.show_pillar_cache()
            assert ret == expected

            expected = {
                "test-host": {
                    "this": "one",
                    "that": "two",
                    "those": ["three", "four", "five"],
                }
            }
            ret = pillar_runner.show_pillar_cache("test-host")
            assert ret == expected

        _EMPTY_CHECK_MINIONS_RETURN = {"minions": [], "missing": []}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(return_value=_EMPTY_CHECK_MINIONS_RETURN),
        ), patch("salt.utils.atomicfile.atomic_open", mock_open()) as atomic_open_mock:
            ret = pillar_runner.show_pillar_cache("fake-host")
            assert ret == {}
